"""A module for fetching retrospective NWM gridded data.

The function ``nwm_retro_grids_to_parquet()`` can be used to fetch and format
two different versions of retrospective NWM data (v2.1 and v3.0) for
specified variables and date ranges, and summarize the grid pixels
intersecting polygons provided in the weights file using an area-weighted mean
approach.

Each version of the NWM data is hosted on `AWS S3
<https://registry.opendata.aws/nwm-archive/>`__ and is stored in Zarr format
(v3.0) or as Kerchunk reference files (v2.1). Several options are included
for fetching the data in chunks, including by week or month which can be
specified using the ``chunk_by`` argument.

The NWM v3.0 retrospective forcing Zarr store has dimensions {time: 385704,
x: 3840, and y: 4608}, with a chunking scheme of {time: 672, x: 350, y: 350}.
Only the data variable chunks that intersect the polygons are read into memory.

The NWM v2.1 retrospective forcing data has the same x-y dimensions but is
fetched using Kerchunk reference files, which point to the original hourly
netcdf files, therefore the data is not chunked in the x-y dimensions. This
means that an entire data variable is read into memory regardless of the
spatial bounds of the polygons being processed.

Care must be taken when choosing a ``chunk_by`` value to minimize the amount
of data transferred over the network.

.. note::
   It is recommended to set the ``chunk_by`` parameter to the largest time
   period ('week' or 'month') that will fit into your systems memory
   given the number of polygons being processed.
"""
from datetime import datetime
from pathlib import Path
from typing import Union, Optional, Tuple, Dict
import logging
import numpy as np

import pandas as pd
import xarray as xr
import fsspec
from pydantic import validate_call, InstanceOf
from geopandas import GeoDataFrame
import dask

from teehr.fetching.const import (
    VALUE_TIME,
    REFERENCE_TIME,
    UNIT_NAME,
    VARIABLE_NAME,
    CONFIGURATION_NAME
)
from teehr.models.fetching.utils import (
    NWMChunkByEnum,
    SupportedNWMRetroVersionsEnum,
    SupportedNWMRetroDomainsEnum,
    TimeseriesTypeEnum
)
from teehr.models.fetching.nwm22_grid import ForcingVariablesEnum
from teehr.fetching.nwm.grid_utils import (
    update_location_id_prefix,
    compute_weighted_average,
    get_nwm_grid_data,
    get_weights_row_col_stats,
    read_and_validate_weights_file
)
from teehr.fetching.utils import (
    write_timeseries_parquet_file,
    get_dataset,
    get_period_start_end_times,
    create_periods_based_on_chunksize
)
from teehr.fetching.const import CONUS_NWM_WKT
from teehr.fetching.nwm.retrospective_points import (
    format_grouped_filename,
    validate_retrospective_start_end_date,
)
from teehr.utilities.generate_weights import generate_weights_file


logger = logging.getLogger(__name__)


def get_nwm21_retro_grid_data(
    var_da: xr.DataArray,
    row_min: int,
    col_min: int,
    row_max: int,
    col_max: int
):
    """Read a subset nwm21 retro grid data into memory from row/col bounds."""
    logger.debug("Getting the nwm21 retro grid data")
    grid_values = var_da.isel(
        west_east=slice(col_min, col_max+1),
        south_north=slice(row_min, row_max+1)
    ).values
    return grid_values


def process_nwm30_retro_group(
    da_i: xr.DataArray,
    weights_filepath: str,
    variable_name: str,
    nwm_version: str,
    location_id_prefix: Union[str, None],
    variable_mapper: Dict[str, Dict[str, str]]
):
    """Compute the weighted average for a chunk of NWM v3.0 data.

    Pixel weights for each zone are defined in weights_df,
    and the output is saved to parquet files.
    """
    logger.debug("Processing NWM v3.0 retro grid data chunk.")
    weights_df = read_and_validate_weights_file(weights_filepath)

    weights_bounds = get_weights_row_col_stats(weights_df)

    grid_arr = get_nwm_grid_data(
        da_i,
        weights_bounds["row_min"],
        weights_bounds["col_min"],
        weights_bounds["row_max"],
        weights_bounds["col_max"]
    )

    hourly_dfs = []
    for i, time in enumerate(da_i.time.values):
        grid_values = grid_arr[
            i,
            weights_bounds["rows_norm"],
            weights_bounds["cols_norm"]
        ]
        df = compute_weighted_average(grid_values, weights_df)
        df.loc[:, VALUE_TIME] = pd.to_datetime(time)
        hourly_dfs.append(df)

    nwm_units = da_i.attrs["units"]
    chunk_df = pd.concat(hourly_dfs)
    if not variable_mapper:
        chunk_df.loc[:, UNIT_NAME] = nwm_units
        chunk_df.loc[:, VARIABLE_NAME] = variable_name.value
    else:
        chunk_df.loc[:, UNIT_NAME] = variable_mapper[UNIT_NAME].\
            get(nwm_units, nwm_units)
        chunk_df.loc[:, VARIABLE_NAME] = variable_mapper[VARIABLE_NAME].\
            get(variable_name, variable_name)

    chunk_df.loc[:, REFERENCE_TIME] = np.nan
    chunk_df.loc[:, CONFIGURATION_NAME] = f"{nwm_version}_retrospective"

    if location_id_prefix:
        chunk_df = update_location_id_prefix(chunk_df, location_id_prefix)

    return chunk_df


def construct_nwm21_json_paths(
    start_date: Union[str, datetime],
    end_date: Union[str, datetime]
):
    """Construct the remote paths for NWM v2.1 json files as a dataframe."""
    logger.debug("Constructing NWM v2.1 json paths.")
    base_path = (
        "s3://ciroh-nwm-zarr-retrospective-data-copy/"
        "noaa-nwm-retrospective-2-1-zarr-pds/forcing"
    )
    date_rng = pd.date_range(start_date, end_date, freq="h")
    dates = []
    paths = []
    for dt in date_rng:
        dates.append(dt)
        paths.append(
            f"{base_path}/{dt.year}/{dt.year}{dt.month:02d}"
            f"{dt.day:02d}{dt.hour:02d}.LDASIN_DOMAIN1.json"
        )
    paths_df = pd.DataFrame({"datetime": dates, "filepath": paths})
    return paths_df


@dask.delayed
def process_single_nwm21_retro_grid_file(
    row: Tuple,
    variable_name: str,
    weights_filepath: str,
    ignore_missing_file: bool,
    nwm_version: str,
    location_id_prefix: Union[str, None],
    variable_mapper: Dict[str, Dict[str, str]]
):
    """Compute the zonal mean for a single json reference file.

    Results are formatted to a dataframe using the TEEHR data model.
    """
    ds = get_dataset(
        row.filepath,
        ignore_missing_file,
        target_options={'anon': True}
    )
    if not ds:
        return None

    nwm_units = ds[variable_name].attrs["units"]

    value_time = row.datetime
    da = ds[variable_name].isel(Time=0)

    weights_df = read_and_validate_weights_file(weights_filepath)

    weights_bounds = get_weights_row_col_stats(weights_df)

    grid_arr = get_nwm21_retro_grid_data(
        da,
        weights_bounds["row_min"],
        weights_bounds["col_min"],
        weights_bounds["row_max"],
        weights_bounds["col_max"]
    )
    grid_values = grid_arr[
        weights_bounds["rows_norm"],
        weights_bounds["cols_norm"]
    ]

    # Calculate mean areal of selected variable
    df = compute_weighted_average(grid_values, weights_df)

    if not variable_mapper:
        df.loc[:, UNIT_NAME] = nwm_units
        df.loc[:, VARIABLE_NAME] = variable_name.value
    else:
        df.loc[:, UNIT_NAME] = variable_mapper[UNIT_NAME].\
            get(nwm_units, nwm_units)
        df.loc[:, VARIABLE_NAME] = variable_mapper[VARIABLE_NAME].\
            get(variable_name, variable_name)

    df.loc[:, VALUE_TIME] = value_time
    df.loc[:, REFERENCE_TIME] = np.nan
    df.loc[:, CONFIGURATION_NAME] = f"{nwm_version}_retrospective"

    if location_id_prefix:
        df = update_location_id_prefix(df, location_id_prefix)

    return df


@validate_call(config=dict(arbitrary_types_allowed=True))
def nwm_retro_grids_to_parquet(
    nwm_version: SupportedNWMRetroVersionsEnum,
    variable_name: ForcingVariablesEnum,
    zonal_weights_filepath: Union[str, Path],
    start_date: Union[str, datetime, pd.Timestamp],
    end_date: Union[str, datetime, pd.Timestamp],
    output_parquet_dir: Union[str, Path],
    chunk_by: Union[NWMChunkByEnum, None] = None,
    overwrite_output: Optional[bool] = False,
    domain: Optional[SupportedNWMRetroDomainsEnum] = "CONUS",
    location_id_prefix: Optional[Union[str, None]] = None,
    variable_mapper: Dict[str, Dict[str, str]] = None,
    timeseries_type: TimeseriesTypeEnum = "primary",
    calculate_zonal_weights: bool = False,
    zone_polygons: Optional[Union[Path, str, InstanceOf[GeoDataFrame]]] = None,
    unique_zone_id: Optional[str] = None,
):
    """Compute the weighted average for NWM v2.1 or v3.0 gridded data.

    Pixel values are summarized to zones based on a pre-computed
    zonal weights file, and the output is saved to parquet files.

    All dates and times within the files and in the file names are in UTC.

    Parameters
    ----------
    nwm_version : SupportedNWMRetroVersionsEnum
        NWM retrospective version to fetch.
        Currently `nwm21` and `nwm30` supported.
    variable_name : str
        Name of the NWM forcing data variable to download.
        (e.g., "PRECIP", "PSFC", "Q2D", ...).
    zonal_weights_filepath : str,
        Path to the array containing fraction of pixel overlap
        for each zone. The values in the location_id field from
        the zonal weights file are used in the output of this function.
    start_date : Union[str, datetime, pd.Timestamp]
        Date to begin data ingest.
        Str formats can include YYYY-MM-DD or MM/DD/YYYY.
        Rounds down to beginning of day.
    end_date : Union[str, datetime, pd.Timestamp],
        Last date to fetch.  Rounds up to end of day.
        Str formats can include YYYY-MM-DD or MM/DD/YYYY.
    output_parquet_dir : Union[str, Path],
        Directory where output will be saved.
    chunk_by : Union[NWMChunkByEnum, None] = None,
        If None (default) saves all timeseries to a single file, otherwise
        the data is processed using the specified parameter.
        Can be: 'week' or 'month' for gridded data.
    overwrite_output : bool = False,
        Whether output should overwrite files if they exist.  Default is False.
    domain : str = "CONUS"
        Geographical domain when NWM version is v3.0.
        Acceptable values are "Alaska", "CONUS" (default), "Hawaii", and "PR".
        Only used when NWM version equals v3.0.
    location_id_prefix : Union[str, None]
        Optional location ID prefix to add (prepend) or replace.
    variable_mapper : Dict[str, Dict[str, str]]
        Dictionary mapping NWM variable and unit names to TEEHR variable
        and unit names.
    calculate_zonal_weights : bool
        Flag to calculate zonal weights.
    zone_polygons : Union[Path, str, InstanceOf[GeoDataFrame]]
        Path to the polygons file or a GeoDataFrame.
    unique_zone_id : Optional[str]
        Name of the field in the zone polygon file containing unique IDs.

    Notes
    -----
    The location_id values in the zonal weights file are used as
    location ids in the output of this function, unless a prefix is specified
    which will be prepended to the location_id values if none exists, or it
    will replace the existing prefix. It is assumed that the location_id
    follows the pattern '[prefix]-[unique id]'.
    """
    logger.info(
        f"Fetching NWM retrospective grid data, version: {nwm_version}."
    )

    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)

    validate_retrospective_start_end_date(nwm_version, start_date, end_date)

    # Include the entirety of the specified end day
    end_date = end_date.to_period(freq="D").end_time

    output_dir = Path(output_parquet_dir)
    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    if nwm_version == SupportedNWMRetroVersionsEnum.nwm21:

        # Construct Kerchunk-json paths within the selected time
        nwm21_paths = construct_nwm21_json_paths(start_date, end_date)

        # If specified, generate zonal weights file here.
        if calculate_zonal_weights:
            if zone_polygons is None:
                raise ValueError(
                    "The zone polygons must be provided"
                    " to calculate zonal weights. Can be a GeoDataFame"
                    " or a filepath."
                )

            # Get a single timestep to use as a template grid.
            template_ds = get_dataset(
                nwm21_paths.filepath[0],
                ignore_missing_file=False,
                target_options={'anon': True}
            )

            # Get spatial information from the zarr store. (limited data)
            tmp_s3_zarr_url = "s3://noaa-nwm-retrospective-2-1-zarr-pds/precip.zarr"
            tmp_ds = xr.open_zarr(fsspec.get_mapper(tmp_s3_zarr_url, anon=True))
            nwm21_crs = tmp_ds.crs.attrs["esri_pe_string"]
            x_dim = tmp_ds.x.values
            y_dim = tmp_ds.y.values

            # Need to rename x and y dimensions in the template dataset.
            template_ds["west_east"] = x_dim
            template_ds["south_north"] = y_dim
            template_ds = template_ds.rename(
                {"west_east": "x", "south_north": "y"}
            )

            generate_weights_file(
                zone_polygons=zone_polygons,
                template_dataset=template_ds,
                variable_name=variable_name,
                crs_wkt=nwm21_crs,
                output_weights_filepath=zonal_weights_filepath,
                location_id_prefix=location_id_prefix,
                unique_zone_id=unique_zone_id
            )

        if chunk_by in ["year"]:
            raise ValueError(
                f"Chunkby '{chunk_by}' is not implemented for gridded data"
            )

        periods = create_periods_based_on_chunksize(
            start_date=start_date,
            end_date=end_date,
            chunk_by=chunk_by
        )

        for period in periods:
            if period is not None:
                dts = get_period_start_end_times(
                    period=period,
                    start_date=start_date,
                    end_date=end_date
                )
                df = nwm21_paths[nwm21_paths.datetime.between(
                    dts["start_dt"], dts["end_dt"]
                )].copy()
            else:
                df = nwm21_paths.copy()

            # Process this chunk using dask delayed
            results = []
            for row in df.itertuples():
                results.append(
                    process_single_nwm21_retro_grid_file(
                        row=row,
                        variable_name=variable_name,
                        weights_filepath=zonal_weights_filepath,
                        ignore_missing_file=False,
                        nwm_version=nwm_version,
                        location_id_prefix=location_id_prefix,
                        variable_mapper=variable_mapper
                    )
                )
            output = dask.compute(*results)

            output = [df for df in output if df is not None]
            if len(output) == 0:
                raise FileNotFoundError("No NWM files for specified input"
                                        "configuration were found in GCS!")
            chunk_df = pd.concat(output)

            start = df.datetime.min().strftime("%Y%m%d")
            end = df.datetime.max().strftime("%Y%m%d")
            if start == end:
                output_filename = Path(output_parquet_dir, f"{start}.parquet")
            else:
                output_filename = Path(
                    output_parquet_dir,
                    f"{start}_{end}.parquet"
                )

            write_timeseries_parquet_file(
                filepath=output_filename,
                overwrite_output=overwrite_output,
                data=chunk_df,
                timeseries_type=timeseries_type
            )

    if nwm_version == SupportedNWMRetroVersionsEnum.nwm30:

        # Construct the path to the zarr store
        if variable_name == "RAINRATE":
            zarr_name = "precip"
        else:
            zarr_name = variable_name.lower()
        s3_zarr_url = (
            f"s3://noaa-nwm-retrospective-3-0-pds/{domain}/"
            f"zarr/forcing/{zarr_name}.zarr"
        )

        ds = xr.open_zarr(
            fsspec.get_mapper(s3_zarr_url, anon=True),
            chunks={}, consolidated=True
        ).sel(time=slice(start_date, end_date))

        # If specified, generate zonal weights file here.
        if calculate_zonal_weights:
            if zone_polygons is None:
                raise ValueError(
                    "The zone polygons must be provided"
                    " to calculate zonal weights. Can be a GeoDataFame"
                    " or a filepath."
                )

            generate_weights_file(
                zone_polygons=zone_polygons,
                template_dataset=ds,
                variable_name=variable_name,
                crs_wkt=ds[variable_name].attrs["esri_pe_string"],
                output_weights_filepath=zonal_weights_filepath,
                location_id_prefix=location_id_prefix,
                unique_zone_id=unique_zone_id
            )

        if chunk_by in ["year", "location_id"]:
            raise ValueError(
                f"Chunkby '{chunk_by}' is not implemented for gridded data"
            )

        periods = create_periods_based_on_chunksize(
            start_date=start_date,
            end_date=end_date,
            chunk_by=chunk_by
        )

        var_da = ds[variable_name]
        for period in periods:

            if period is not None:
                dts = get_period_start_end_times(
                    period=period,
                    start_date=start_date,
                    end_date=end_date
                )
            else:
                dts = {"start_dt": start_date, "end_dt": end_date}

            da_i = var_da.sel(time=slice(dts["start_dt"], dts["end_dt"]))

            chunk_df = process_nwm30_retro_group(
                da_i=da_i,
                weights_filepath=zonal_weights_filepath,
                variable_name=variable_name,
                nwm_version=nwm_version,
                location_id_prefix=location_id_prefix,
                variable_mapper=variable_mapper
            )

            fname = format_grouped_filename(da_i)
            output_filename = Path(
                output_parquet_dir,
                fname
            )

            write_timeseries_parquet_file(
                filepath=output_filename,
                overwrite_output=overwrite_output,
                data=chunk_df,
                timeseries_type=timeseries_type
            )



"""Classes for probabilistic metric calculation methods."""
from typing import Callable, List, Union, Dict

from pydantic import Field

from teehr.models.metrics.basemodels import ProbabilisticBasemodel
from teehr.metrics import probabilistic_funcs
import teehr.models.metrics.metric_attributes as tma
from teehr.models.metrics.basemodels import CRPSEstimators
from teehr.models.str_enum import StrEnum
from teehr.models.metrics.basemodels import TransformEnum


class CRPS(ProbabilisticBasemodel):
    """Continous Ranked Probability Score - Ensemble.

    Parameters
    ----------
    estimator : str
        CRPS estimator, can be ("pwm", "nrg", or "fair"). Default is "pwm".
    backend : str
        The backend to use, by default "numba". Can be ("numba" or "numpy").
    summary_func : Callable
        The function to apply to the results, by default np.mean.
    output_field_name : str
        The output field name, by default "mean_crps_ensemble".
    func : Callable
        The function to apply to the data, by default
        :func:`probabilistic_funcs.ensemble_crps`.
    input_field_names : Union[str, StrEnum, List[Union[str, StrEnum]]]
        The input field names, by default
        ["primary_value", "secondary_value", "value_time"].
    attrs : Dict
        The static attributes for the metric.
    """

    estimator: CRPSEstimators = Field(default="pwm")
    transform: TransformEnum = Field(default=None)
    backend: str = Field(default="numba")
    output_field_name: str = Field(default="mean_crps_ensemble")
    func: Callable = Field(probabilistic_funcs.create_crps_func, frozen=True)
    summary_func: Union[Callable, None] = Field(default=None)
    input_field_names: Union[str, StrEnum, List[Union[str, StrEnum]]] = Field(
        default=["primary_value", "secondary_value", "value_time"]
    )
    attrs: Dict = Field(default=tma.CRPS_ENSEMBLE_ATTRS, frozen=True)


class ProbabilisticMetrics:
    """Define and customize probalistic metrics.

    Notes
    -----
    Probabilistic metrics compare a value against a distribution of predicted
    values, such as ensemble forecasts. Available probabilistic metrics include:

    - CRPS (Continuous Ranked Probability Score)
    """

    CRPS = CRPS



"""Module for importing location crosswalks from a file."""
from typing import Union
from pathlib import Path
from teehr.loading.utils import read_and_convert_netcdf_to_df
import teehr.models.pandera_dataframe_schemas as schemas
from teehr.loading.utils import merge_field_mappings
import pandas as pd
import teehr.models.pandera_dataframe_schemas as schemas

import logging

logger = logging.getLogger(__name__)


def convert_single_location_crosswalks(
    in_filepath: Union[str, Path],
    out_filepath: Union[str, Path],
    field_mapping: dict,
    **kwargs
):
    """Convert location_crosswalks data to parquet format.

    Parameters
    ----------
    in_filepath : Union[str, Path]
        The input file path.
    out_filepath : Union[str, Path]
        The output file path.
    field_mapping : dict
        A dictionary mapping input fields to output fields.
        format: {input_field: output_field}
    **kwargs
        Additional keyword arguments are passed to
            pd.read_csv() or pd.read_parquet().
    """
    in_filepath = Path(in_filepath)
    out_filepath = Path(out_filepath)

    logger.info(f"Converting location crosswalks data from: {in_filepath}")

    if in_filepath.suffix == ".parquet":
        # read and convert parquet file
        location_crosswalks = pd.read_parquet(in_filepath, **kwargs)
    elif in_filepath.suffix == ".csv":
        # read and convert csv file
        location_crosswalks = pd.read_csv(in_filepath, **kwargs)
    elif in_filepath.suffix == ".nc":
        # read and convert netcdf file
        location_crosswalks = read_and_convert_netcdf_to_df(
            in_filepath,
            field_mapping,
            **kwargs
        )
    else:
        raise ValueError("Unsupported file type.")

    # rename fields if field_mapping provided
    location_crosswalks.rename(columns=field_mapping, inplace=True)

    # validate schema
    schema = schemas.location_crosswalks_schema(type="pandas")
    validated_df = schema.validate(location_crosswalks)

    # write to parquet
    logger.info(f"Writing location crosswalks to: {out_filepath}")
    out_filepath.parent.mkdir(parents=True, exist_ok=True)
    validated_df.to_parquet(out_filepath)


def convert_location_crosswalks(
    in_path: Union[str, Path],
    out_dirpath: Union[str, Path],
    field_mapping: dict = None,
    pattern: str = "**/*.parquet",
    **kwargs
):
    """Convert crosswalk data to parquet format.

    Parameters
    ----------
    in_path : Union[str, Path]
        The input file or directory path.
    out_dirpath : Union[str, Path]
        The output directory path.
    field_mapping : dict, optional
        A dictionary mapping input fields to output fields.
        format: {input_field: output_field}
    pattern : str, optional (default: "**/*.parquet")
        The pattern to match files.
    **kwargs
        Additional keyword arguments are passed to
            pd.read_csv() or pd.read_parquet().
    """
    in_path = Path(in_path)
    out_dirpath = Path(out_dirpath)
    logger.info(f"Converting crosswalks data: {in_path}")

    default_field_mapping = {}
    for field in schemas.location_crosswalks_schema(type="pandas").columns.keys():
        if field not in default_field_mapping.values():
            default_field_mapping[field] = field

    if field_mapping:
        logger.debug("Merging user field_mapping with default field mapping.")
        field_mapping = merge_field_mappings(
            default_field_mapping,
            field_mapping
        )
    else:
        logger.debug("Using default field mapping.")
        field_mapping = default_field_mapping

    files_converted = 0
    if in_path.is_dir():
        # recursively convert all files in directory
        logger.info(f"Recursively converting all files in {in_path}/{pattern}")
        for in_filepath in in_path.glob(f"{pattern}"):
            relative_name = in_filepath.relative_to(in_path)
            out_filepath = Path(out_dirpath, relative_name)
            out_filepath = out_filepath.with_suffix(".parquet")
            convert_single_location_crosswalks(
                in_filepath,
                out_filepath,
                field_mapping,
                **kwargs
            )
            files_converted += 1
    else:
        out_filepath = Path(out_dirpath, in_path.name)
        out_filepath = out_filepath.with_suffix(".parquet")
        convert_single_location_crosswalks(
            in_path,
            out_filepath,
            field_mapping,
            **kwargs
        )
        files_converted += 1
    logger.info(f"Converted {files_converted} files.")



"""This is a StrEnum class to work around breaking changes in Python 3.11."""
try:
    # breaking change introduced in python 3.11
    from enum import StrEnum
except ImportError:  # pragma: no cover
    from enum import Enum  # pragma: no cover

    class StrEnum(str, Enum):  # pragma: no cover
        """Enum with string values."""

        pass  # pragma: no cover



"""Module for fetchning and processing NWM point data."""
from typing import Union, Optional, List, Dict, Annotated
from datetime import datetime
from pathlib import Path
import logging

from pydantic import validate_call, Field

from teehr.fetching.nwm.point_utils import (
    fetch_and_format_nwm_points,
)
from teehr.fetching.utils import (
    generate_json_paths,
    build_remote_nwm_filelist,
    validate_operational_start_end_date,
    start_on_z_hour,
    end_on_z_hour
)
from teehr.models.fetching.utils import (
    SupportedNWMOperationalVersionsEnum,
    SupportedNWMDataSourcesEnum,
    SupportedKerchunkMethod,
    TimeseriesTypeEnum
)
from teehr.fetching.const import (
    NWM12_ANALYSIS_CONFIG,
    NWM20_ANALYSIS_CONFIG,
    NWM22_ANALYSIS_CONFIG,
    NWM30_ANALYSIS_CONFIG,
)

logger = logging.getLogger(__name__)


@validate_call()
def nwm_to_parquet(
    configuration: str,
    output_type: str,
    variable_name: str,
    start_date: Union[str, datetime],
    ingest_days: int,
    location_ids: List[int],
    json_dir: Union[str, Path],
    output_parquet_dir: Union[str, Path],
    nwm_version: SupportedNWMOperationalVersionsEnum,
    data_source: Optional[SupportedNWMDataSourcesEnum] = "GCS",
    kerchunk_method: Optional[SupportedKerchunkMethod] = "local",
    prioritize_analysis_valid_time: Optional[bool] = False,
    t_minus_hours: Optional[List[int]] = None,
    process_by_z_hour: Optional[bool] = True,
    stepsize: Optional[int] = 100,
    ignore_missing_file: Optional[bool] = True,
    overwrite_output: Optional[bool] = False,
    variable_mapper: Dict[str, Dict[str, str]] = None,
    timeseries_type: TimeseriesTypeEnum = "secondary",
    starting_z_hour: Optional[Annotated[int, Field(ge=0, le=23)]] = None,
    ending_z_hour: Optional[Annotated[int, Field(ge=0, le=23)]] = None
):
    """Fetch NWM point data and save as a Parquet file in TEEHR format.

    Parameters
    ----------
    configuration : str
        NWM forecast category.
        (e.g., "analysis_assim", "short_range", ...).
    output_type : str
        Output component of the configuration.
        (e.g., "channel_rt", "reservoir", ...).
    variable_name : str
        Name of the NWM data variable to download.
        (e.g., "streamflow", "velocity", ...).
    start_date : str or datetime
        Date to begin data ingest.
        Str formats can include YYYY-MM-DD or MM/DD/YYYY.
    ingest_days : int
        Number of days to ingest data after start date.
    location_ids : List[int]
        Array specifying NWM IDs of interest.
    json_dir : str
        Directory path for saving json reference files.
    output_parquet_dir : str
        Path to the directory for the final parquet files.
    nwm_version : SupportedNWMOperationalVersionsEnum
        The NWM operational version.
        "nwm12", "nwm20", "nwm21", "nwm22", or "nwm30".
        Note that there is no change in NWM configuration between
        version 2.1 and 2.2, and they are treated as the same version.
        They are both allowed here for convenience.

        Availability of each version:

        - v1.2: 2018-09-17 - 2019-06-18
        - v2.0: 2019-06-19 - 2021-04-19
        - v2.1/2.2: 2021-04-20 - 2023-09-18
        - v3.0: 2023-09-19 - present
    data_source : Optional[SupportedNWMDataSourcesEnum]
        Specifies the remote location from which to fetch the data
        "GCS" (default), "NOMADS", or "DSTOR"
        Currently only "GCS" is implemented.
    kerchunk_method : Optional[SupportedKerchunkMethod]
        When data_source = "GCS", specifies the preference in creating Kerchunk
        reference json files. "local" (default) will create new json files from
        netcdf files in GCS and save to a local directory if they do not already
        exist locally, in which case the creation is skipped. "remote" - read the
        CIROH pre-generated jsons from s3, ignoring any that are unavailable.
        "auto" - read the CIROH pre-generated jsons from s3, and create any that
        are unavailable, storing locally.
    prioritize_analysis_valid_time : Optional[bool]
        A boolean flag that determines the method of fetching analysis data.
        When False (default), all hours of the reference time are included in the
        output. When True, only the hours within t_minus_hours are included.
    t_minus_hours : Optional[List[int]]
        Specifies the look-back hours to include if an assimilation
        configuration is specified.
    process_by_z_hour : Optional[bool]
        A boolean flag that determines the method of grouping files
        for processing. The default is True, which groups by day and z_hour.
        False groups files sequentially into chunks, whose size is determined
        by stepsize. This allows users to process more data potentially more
        efficiently, but runs to risk of splitting up forecasts into separate
        output files.
    stepsize : Optional[int]
        The number of json files to process at one time. Used if
        process_by_z_hour is set to False. Default value is 100. Larger values
        can result in greater efficiency but require more memory.
    ignore_missing_file : Optional[bool]
        Flag specifying whether or not to fail if a missing NWM file is
        encountered. True = skip and continue; False = fail.
    overwrite_output : Optional[bool]
        Flag specifying whether or not to overwrite output files if they
        already exist.  True = overwrite; False = fail.
    timeseries_type : str
        Whether to consider as the "primary" or "secondary" timeseries.
        Default is "secondary".
    starting_z_hour : Optional[int]
        The starting z_hour to include in the output. If None, all z_hours
        are included for the first day. Default is None. Must be between 0 and 23.
    ending_z_hour : Optional[int]
        The ending z_hour to include in the output. If None, all z_hours
        are included for the last day. Default is None. Must be between 0 and 23.

    Notes
    -----
    The NWM configuration variables, including configuration, output_type, and
    variable_name are stored as pydantic models in point_config_models.py

    Forecast and assimilation data is grouped and saved one file per reference
    time, using the file name convention "YYYYMMDDTHH".  The tabular output
    parquet files follow the timeseries data model described in the
    :ref:`data model <data_model>`.

    All dates and times within the files and in the file names are in UTC.

    Examples
    --------
    Here we fetch operational streamflow forecasts for NWM v2.2 from GCS, and
    save the output in the TEEHR :ref:`data model <data_model>` format.

    Import the necessary module.

    >>> import teehr.fetching.nwm.nwm_points as tlp

    Specify the input variables.

    >>> CONFIGURATION = "short_range"
    >>> OUTPUT_TYPE = "channel_rt"
    >>> VARIABLE_NAME = "streamflow"
    >>> START_DATE = "2023-03-18"
    >>> INGEST_DAYS = 1
    >>> JSON_DIR = Path(Path.home(), "temp/parquet/jsons/")
    >>> OUTPUT_DIR = Path(Path.home(), "temp/parquet")
    >>> NWM_VERSION = "nwm22"
    >>> DATA_SOURCE = "GCS"
    >>> KERCHUNK_METHOD = "auto"
    >>> T_MINUS = [0, 1, 2]
    >>> IGNORE_MISSING_FILE = True
    >>> OVERWRITE_OUTPUT = True
    >>> PROCESS_BY_Z_HOUR = True
    >>> STEPSIZE = 100
    >>> LOCATION_IDS = [7086109,  7040481,  7053819]

    Fetch and format the data, writing to the specified directory.

    >>> tlp.nwm_to_parquet(
    >>>     nwm_configuration=CONFIGURATION,
    >>>     output_type=OUTPUT_TYPE,
    >>>     variable_name=VARIABLE_NAME,
    >>>     start_date=START_DATE,
    >>>     ingest_days=INGEST_DAYS,
    >>>     location_ids=LOCATION_IDS,
    >>>     json_dir=JSON_DIR,
    >>>     output_parquet_dir=OUTPUT_DIR,
    >>>     nwm_version=NWM_VERSION,
    >>>     data_source=DATA_SOURCE,
    >>>     kerchunk_method=KERCHUNK_METHOD,
    >>>     t_minus_hours=T_MINUS,
    >>>     process_by_z_hour=PROCESS_BY_Z_HOUR,
    >>>     stepsize=STEPSIZE,
    >>>     ignore_missing_file=IGNORE_MISSING_FILE,
    >>>     overwrite_output=OVERWRITE_OUTPUT,
    >>> )
    """ # noqa
    logger.info(f"Fetching {configuration} data. Version: {nwm_version}")

    # Import appropriate config model and dicts based on NWM version
    if nwm_version == SupportedNWMOperationalVersionsEnum.nwm12:
        from teehr.models.fetching.nwm12_point import PointConfigurationModel
        analysis_config_dict = NWM12_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm20:
        from teehr.models.fetching.nwm20_point import PointConfigurationModel
        analysis_config_dict = NWM20_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm21:
        from teehr.models.fetching.nwm22_point import PointConfigurationModel
        analysis_config_dict = NWM22_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm22:
        from teehr.models.fetching.nwm22_point import PointConfigurationModel
        analysis_config_dict = NWM22_ANALYSIS_CONFIG
    elif nwm_version == SupportedNWMOperationalVersionsEnum.nwm30:
        from teehr.models.fetching.nwm30_point import PointConfigurationModel
        analysis_config_dict = NWM30_ANALYSIS_CONFIG
    else:
        raise ValueError(
            "nwm_version must equal "
            "'nwm12', 'nwm20', 'nwm21', 'nwm22' or 'nwm30'"
        )

    # Parse input parameters to validate configuration
    vars = {
        "configuration": configuration,
        configuration: {
            "output_type": output_type,
            output_type: variable_name,
        },
    }
    cm = PointConfigurationModel.model_validate(vars)
    configuration = cm.configuration.name
    forecast_obj = getattr(cm, configuration)
    output_type = forecast_obj.output_type.name
    variable_name = getattr(forecast_obj, output_type).name

    # Check data_source
    if data_source == SupportedNWMDataSourcesEnum.NOMADS:
        # TODO
        raise ValueError("Fetching from NOMADS is not yet implemented")
    elif data_source == SupportedNWMDataSourcesEnum.DSTOR:
        # TODO
        raise ValueError("Fetching from DSTOR is not yet implemented")
    else:

        # Make sure start/end dates work with specified NWM version
        validate_operational_start_end_date(
            nwm_version,
            start_date,
            ingest_days
        )

        # Build paths to netcdf files on GCS
        gcs_component_paths = build_remote_nwm_filelist(
            configuration,
            output_type,
            start_date,
            ingest_days,
            analysis_config_dict,
            t_minus_hours,
            ignore_missing_file,
            prioritize_analysis_valid_time
        )

        if starting_z_hour is not None:
            gcs_component_paths = start_on_z_hour(
                start_date=start_date,
                start_z_hour=starting_z_hour,
                gcs_component_paths=gcs_component_paths
            )

        if ending_z_hour is not None:
            gcs_component_paths = end_on_z_hour(
                start_date=start_date,
                ingest_days=ingest_days,
                end_z_hour=ending_z_hour,
                gcs_component_paths=gcs_component_paths
            )

        # Create paths to local and/or remote kerchunk jsons
        json_paths = generate_json_paths(
            kerchunk_method,
            gcs_component_paths,
            json_dir,
            ignore_missing_file
        )

        # Fetch the data, saving to parquet files based on TEEHR data model
        fetch_and_format_nwm_points(
            json_paths,
            location_ids,
            configuration,
            variable_name,
            output_parquet_dir,
            process_by_z_hour,
            stepsize,
            ignore_missing_file,
            overwrite_output,
            nwm_version,
            variable_mapper,
            timeseries_type
        )



"""Module to fetch test data."""

import requests


files = [
    {
        "name": "gages.geojson",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/v0_3_test_study/geo/gages.geojson"
    },
    {
        "name": "test_short_obs.csv",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/v0_3_test_study/timeseries/test_short_obs.csv"
    },
    {
        "name": "test_short_fcast.parquet",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/v0_3_test_study/timeseries/test_short_fcast.parquet"
    },
    {
        "name": "crosswalk.csv",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/v0_3_test_study/geo/crosswalk.csv"
    },
    {
        "name": "test_attr_2yr_discharge.csv",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/v0_3_test_study/geo/test_attr_2yr_discharge.csv"
    },
    {
        "name": "test_attr_drainage_area_km2.csv",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/v0_3_test_study/geo/test_attr_drainage_area_km2.csv"
    },
    {
        "name": "test_attr_ecoregion.csv",
        "url": "https://github.com/RTIInternational/teehr/raw/refs/heads/main/tests/data/v0_3_test_study/geo/test_attr_ecoregion.csv"
    }
]

def fetch_and_save_file(url: str, destination: str) -> None:
    """Fetch a file from a URL and save it to a destination."""
    response = requests.get(url)
    response.raise_for_status()  # Ensure we notice bad responses

    with open(destination, 'wb') as f:
        f.write(response.content)


def list_files() -> list:
    return [file["name"] for file in files]


def fetch_file(file_name: str, destination: str) -> None:
    """Fetch a file by name."""
    file = next(file for file in files if file["name"] == file_name)
    fetch_and_save_file(file["url"], destination)





"""Module describing NWM v3.0 point configuration variables."""
from teehr.models.str_enum import StrEnum

from typing import Optional

from pydantic import BaseModel

from teehr.models.fetching.nwm22_point import (
    ShortAndAnalysis,
    ShortAndAnalysisNoDA,
    MediumRangeMem1,
    MediumRangeMem2,
    MediumRangeMem3,
    MediumRangeMem4,
    MediumRangeMem5,
    MediumRangeMem6,
    MediumRangeMem7,
    MediumRangeNoDA,
    LongRangeMem1,
    LongRangeMem2,
    LongRangeMem3,
    LongRangeMem4
)


# POINT CONFIGURATION ENUM: Potential configuration names
class ConfigurationsEnum(StrEnum):
    """ConfigurationsEnum."""
    analysis_assim = "analysis_assim"
    analysis_assim_no_da = "analysis_assim_no_da"
    analysis_assim_extend = "analysis_assim_extend"
    analysis_assim_extend_no_da = "analysis_assim_extend_no_da"
    analysis_assim_long = "analysis_assim_long"
    analysis_assim_long_no_da = "analysis_assim_long_no_da"
    analysis_assim_hawaii = "analysis_assim_hawaii"
    analysis_assim_hawaii_no_da = "analysis_assim_hawaii_no_da"
    analysis_assim_puertorico = "analysis_assim_puertorico"
    analysis_assim_puertorico_no_da = "analysis_assim_puertorico_no_da"
    analysis_assim_alaska = "analysis_assim_alaska"
    analysis_assim_alaska_no_da = "analysis_assim_alaska_no_da"
    analysis_assim_extend_alaska = "analysis_assim_extend_alaska"
    analysis_assim_extend_alaska_no_da = "analysis_assim_extend_alaska_no_da"
    short_range = "short_range"
    short_range_hawaii = "short_range_hawaii"
    short_range_puertorico = "short_range_puertorico"
    short_range_hawaii_no_da = "short_range_hawaii_no_da"
    short_range_puertorico_no_da = "short_range_puertorico_no_da"
    short_range_alaska = "short_range_alaska"
    medium_range_mem1 = "medium_range_mem1"
    medium_range_mem2 = "medium_range_mem2"
    medium_range_mem3 = "medium_range_mem3"
    medium_range_mem4 = "medium_range_mem4"
    medium_range_mem5 = "medium_range_mem5"
    medium_range_mem6 = "medium_range_mem6"
    medium_range_mem7 = "medium_range_mem7"
    medium_range_no_da = "medium_range_no_da"
    medium_range_alaska_mem1 = "medium_range_alaska_mem1"
    medium_range_alaska_mem2 = "medium_range_alaska_mem2"
    medium_range_alaska_mem3 = "medium_range_alaska_mem3"
    medium_range_alaska_mem4 = "medium_range_alaska_mem4"
    medium_range_alaska_mem5 = "medium_range_alaska_mem5"
    medium_range_alaska_mem6 = "medium_range_alaska_mem6"
    medium_range_alaska_no_da = "medium_range_alaska_no_da"
    medium_range_blend = "medium_range_blend"
    medium_range_blend_alaska = "medium_range_blend_alaska"
    long_range_mem1 = "long_range_mem1"
    long_range_mem2 = "long_range_mem2"
    long_range_mem3 = "long_range_mem3"
    long_range_mem4 = "long_range_mem4"


# POINT CONFIGURATION MODEL
class PointConfigurationModel(BaseModel):
    """NWM v3.0 PointConfigurationModel."""
    configuration: ConfigurationsEnum = None
    analysis_assim: Optional[ShortAndAnalysis] = None
    analysis_assim_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_extend: Optional[ShortAndAnalysis] = None
    analysis_assim_extend_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_long: Optional[ShortAndAnalysis] = None
    analysis_assim_long_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_hawaii: Optional[ShortAndAnalysis] = None
    analysis_assim_hawaii_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_puertorico: Optional[ShortAndAnalysis] = None
    analysis_assim_puertorico_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_alaska: Optional[ShortAndAnalysis] = None
    analysis_assim_alaska_no_da: Optional[ShortAndAnalysisNoDA] = None
    analysis_assim_extend_alaska: Optional[ShortAndAnalysis] = None
    analysis_assim_extend_alaska_no_da: Optional[ShortAndAnalysisNoDA] = None
    short_range: Optional[ShortAndAnalysis] = None
    short_range_hawaii: Optional[ShortAndAnalysis] = None
    short_range_puertorico: Optional[ShortAndAnalysis] = None
    short_range_hawaii_no_da: Optional[ShortAndAnalysisNoDA] = None
    short_range_puertorico_no_da: Optional[ShortAndAnalysisNoDA] = None
    short_range_alaska: Optional[ShortAndAnalysis] = None
    medium_range_mem1: Optional[MediumRangeMem1] = None
    medium_range_mem2: Optional[MediumRangeMem2] = None
    medium_range_mem3: Optional[MediumRangeMem3] = None
    medium_range_mem4: Optional[MediumRangeMem4] = None
    medium_range_mem5: Optional[MediumRangeMem5] = None
    medium_range_mem6: Optional[MediumRangeMem6] = None
    medium_range_mem7: Optional[MediumRangeMem7] = None
    medium_range_no_da: Optional[MediumRangeNoDA] = None
    medium_range_alaska_mem1: Optional[MediumRangeMem1] = None
    medium_range_alaska_mem2: Optional[MediumRangeMem2] = None
    medium_range_alaska_mem3: Optional[MediumRangeMem3] = None
    medium_range_alaska_mem4: Optional[MediumRangeMem4] = None
    medium_range_alaska_mem5: Optional[MediumRangeMem5] = None
    medium_range_alaska_mem6: Optional[MediumRangeMem6] = None
    medium_range_alaska_no_da: Optional[ShortAndAnalysisNoDA] = None
    medium_range_blend: Optional[ShortAndAnalysis] = None
    medium_range_blend_alaska: Optional[ShortAndAnalysis] = None
    long_range_mem1: Optional[LongRangeMem1] = None
    long_range_mem2: Optional[LongRangeMem2] = None
    long_range_mem3: Optional[LongRangeMem3] = None
    long_range_mem4: Optional[LongRangeMem4] = None


if __name__ == "__main__":
    # So for example:
    configuration = "short_range_alaska"
    output_type = "channel_rt"
    variable_name = "streamflow"

    # Assemble input parameters
    vars = {
        "configuration": configuration,
        configuration: {
            "output_type": output_type,
            output_type: variable_name,
        },
    }

    # Check input parameters
    cm = PointConfigurationModel.model_validate(vars)

    config = cm.configuration.name
    forecast_obj = getattr(cm, config)
    out_type = forecast_obj.output_type.name
    var_name = getattr(forecast_obj, out_type).name

    pass



"""Module describing NWM v2.0 point data configuration variables."""
from teehr.models.str_enum import StrEnum

from typing import Optional

from pydantic import BaseModel

from teehr.models.fetching.nwm22_point import (
    ShortAndAnalysis,
    LongRangeMem1,
    LongRangeMem2,
    LongRangeMem3,
    LongRangeMem4,
    ChannelRtVariableEnum,
    TerrainRtVariableEnum,
    ReservoirVariableEnum
)


class MediumOutputEnum(StrEnum):
    """MediumOutputEnum."""

    channel_rt = "channel_rt"
    terrain_rt = "terrain_rt"
    reservoir = "reservoir"


class MediumRange(BaseModel):
    """MediumRange."""

    output_type: MediumOutputEnum
    channel_rt: Optional[ChannelRtVariableEnum] = None
    terrain_rt: Optional[TerrainRtVariableEnum] = None
    reservoir: Optional[ReservoirVariableEnum] = None


# POINT CONFIGURATION ENUM: Potential configuration names
class ConfigurationsEnum(StrEnum):
    """ConfigurationsEnum."""

    analysis_assim = "analysis_assim"
    short_range = "short_range"
    medium_range = "medium_range"
    long_range_mem1 = "long_range_mem1"
    long_range_mem2 = "long_range_mem2"
    long_range_mem3 = "long_range_mem3"
    long_range_mem4 = "long_range_mem4"


# POINT CONFIGURATION MODEL
class PointConfigurationModel(BaseModel):
    """NWM v1.2 PointConfigurationModel."""

    configuration: ConfigurationsEnum = None
    analysis_assim: Optional[ShortAndAnalysis] = None
    short_range: Optional[ShortAndAnalysis] = None
    medium_range: Optional[MediumRange] = None
    long_range_mem1: Optional[LongRangeMem1] = None
    long_range_mem2: Optional[LongRangeMem2] = None
    long_range_mem3: Optional[LongRangeMem3] = None
    long_range_mem4: Optional[LongRangeMem4] = None


if __name__ == "__main__":
    # So for example:
    configuration = "medium_range"
    output_type = "channel_rt"
    variable_name = "streamflow"

    # Assemble input parameters
    vars = {
        "configuration": configuration,
        configuration: {
            "output_type": output_type,
            output_type: variable_name,
        },
    }

    # Check input parameters
    cm = PointConfigurationModel.model_validate(vars)

    config = cm.configuration.name
    forecast_obj = getattr(cm, config)
    out_type = forecast_obj.output_type.name
    var_name = getattr(forecast_obj, out_type).name

    pass



"""Primary timeseries table class."""
import teehr.const as const
from teehr.evaluation.tables.timeseries_table import TimeseriesTable
from teehr.models.table_enums import TimeseriesFields
from teehr.loading.timeseries import convert_timeseries
import teehr.models.pandera_dataframe_schemas as schemas
from pathlib import Path
from typing import Union
import logging
from teehr.utils.utils import to_path_or_s3path, remove_dir_if_exists
from teehr.models.table_enums import TableWriteEnum
from teehr.loading.utils import add_or_replace_sdf_column_prefix


logger = logging.getLogger(__name__)


class PrimaryTimeseriesTable(TimeseriesTable):
    """Access methods to primary timeseries table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "primary_timeseries"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.schema_func = schemas.primary_timeseries_schema
        self.partition_by = [
            "configuration_name",
            "variable_name",
        ]

    def field_enum(self) -> TimeseriesFields:
        """Get the timeseries fields enum."""
        fields = self._get_schema("pandas").columns.keys()
        return TimeseriesFields(
            "TimeseriesFields",
            {field: field for field in fields}
        )

    def _get_schema(self, type: str = "pyspark"):
        """Get the primary timeseries schema."""
        if type == "pandas":
            return self.schema_func(type="pandas")

        location_ids = self.ev.locations.distinct_values("id")
        variable_names = self.ev.variables.distinct_values("name")
        configuration_names = self.ev.configurations.distinct_values("name")
        unit_names = self.ev.units.distinct_values("name")
        return self.schema_func(
            location_ids=location_ids,
            variable_names=variable_names,
            configuration_names=configuration_names,
            unit_names=unit_names
        )

    def _load(
        self,
        in_path: Union[Path, str],
        pattern="**/*.parquet",
        field_mapping: dict = None,
        constant_field_values: dict = None,
        location_id_prefix: str = None,
        write_mode: TableWriteEnum = "append",
        **kwargs
    ):
        """Import timeseries helper."""
        cache_dir = Path(
            self.ev.dir_path,
            const.CACHE_DIR,
            const.LOADING_CACHE_DIR,
            const.PRIMARY_TIMESERIES_DIR
        )
        # Clear the cache directory if it exists.
        remove_dir_if_exists(cache_dir)

        convert_timeseries(
            in_path=in_path,
            out_path=cache_dir,
            field_mapping=field_mapping,
            constant_field_values=constant_field_values,
            timeseries_type="primary",
            pattern=pattern,
            **kwargs
        )

        # Read the converted files to Spark DataFrame
        df = self._read_files(cache_dir)

        # Add or replace location_id prefix if provided
        if location_id_prefix:
            df = add_or_replace_sdf_column_prefix(
                sdf=df,
                column_name="location_id",
                prefix=location_id_prefix,
            )

        # Validate using the _validate() method
        validated_df = self._validate(df)

        # Write to the table
        self._write_spark_df(validated_df, write_mode=write_mode)

        # Reload the table
        self._load_table()


"""A module for fetching retrospective NWM point data (i.e., streamflow).

The function ``nwm_retro_to_parquet()`` can be used to fetch and format three
different versions of retrospective NWM data (v2.0, v2.1, and v3.0) for
specified variables, locations, and date ranges.

Each version of the NWM data is hosted on `AWS S3
<https://registry.opendata.aws/nwm-archive/>`__ and is stored in Zarr format.
Several options are included for fetching the data in chunks, including by
week, month, or year, which can be specified using the ``chunk_by`` argument.

Care must be taken when choosing a ``chunk_by`` value to minimize the amount
of data transferred over the network.

The Zarr stores for each version of the NWM data have the same internal
chunking scheme: {"time": 672, "feature_id": 30000}, which defines the minimum
amount of data that can be accessed at once (a single chunk). This means that
if you specify ``chunk_by='week'``, the entire chunk
(672 hours x 30000 locations) will be fetched for each week (128 hours)
falling within that chunk, resulting in redundant data transfer.

While it is currently not possible to avoid all redundant data transfer,
it can be minimized by selecting the largest chunk_by size that can fit into
your computer's memory.

.. note::
   It is recommended to set the ``chunk_by`` parameter to the largest time
   period ('week', 'month', or 'year') that will fit into your systems memory
   given the number of locations being fetched.
"""
import pandas as pd
import xarray as xr
import fsspec
from pydantic import validate_call
import numpy as np

from datetime import datetime

from pathlib import Path
from typing import Union, List, Optional, Dict
import logging

from teehr.fetching.const import (
    VALUE,
    VALUE_TIME,
    REFERENCE_TIME,
    LOCATION_ID,
    UNIT_NAME,
    VARIABLE_NAME,
    CONFIGURATION_NAME,
    MEMBER
)
from teehr.models.fetching.utils import (
    NWMChunkByEnum,
    SupportedNWMRetroVersionsEnum,
    SupportedNWMRetroDomainsEnum,
    ChannelRtRetroVariableEnum,
    TimeseriesTypeEnum
)
from teehr.fetching.utils import (
    write_timeseries_parquet_file,
    get_period_start_end_times,
    create_periods_based_on_chunksize
)

NWM20_MIN_DATE = datetime(1993, 1, 1)
NWM20_MAX_DATE = datetime(2018, 12, 31, 23)
NWM21_MIN_DATE = pd.Timestamp(1979, 1, 1)
NWM21_MAX_DATE = pd.Timestamp(2020, 12, 31, 23)
NWM30_MIN_DATE = pd.Timestamp(1979, 2, 1, 1)
NWM30_MAX_DATE = pd.Timestamp(2023, 1, 31, 23)

logger = logging.getLogger(__name__)


def validate_retrospective_start_end_date(
    nwm_version: SupportedNWMRetroVersionsEnum,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime]
):
    """Validate the start and end dates by NWM version."""
    if nwm_version == SupportedNWMRetroVersionsEnum.nwm20:
        if end_date <= start_date:
            raise ValueError("start_date must be before end_date")

        if start_date < NWM21_MIN_DATE:
            raise ValueError(
                f"start_date must be on or after {NWM20_MIN_DATE}"
            )

        if end_date > NWM21_MAX_DATE:
            raise ValueError(f"end_date must be on or before {NWM20_MAX_DATE}")

    elif nwm_version == SupportedNWMRetroVersionsEnum.nwm21:
        if end_date <= start_date:
            raise ValueError("start_date must be before end_date")

        if start_date < NWM21_MIN_DATE:
            raise ValueError(
                f"start_date must be on or after {NWM21_MIN_DATE}"
            )

        if end_date > NWM21_MAX_DATE:
            raise ValueError(f"end_date must be on or before {NWM21_MAX_DATE}")
    elif nwm_version == SupportedNWMRetroVersionsEnum.nwm30:
        if end_date <= start_date:
            raise ValueError("start_date must be before end_date")

        if start_date < NWM30_MIN_DATE:
            raise ValueError(
                f"start_date must be on or after {NWM30_MIN_DATE}"
            )

        if end_date > NWM30_MAX_DATE:
            raise ValueError(f"end_date must be on or before {NWM30_MAX_DATE}")
    else:
        raise ValueError(f"unsupported NWM version {nwm_version}")


def da_to_df(
        nwm_version: SupportedNWMRetroVersionsEnum,
        da: xr.DataArray,
        variable_mapper: Dict[str, Dict[str, str]]
) -> pd.DataFrame:
    """Format NWM retrospective data to TEEHR format."""
    logger.debug("Converting DataArray to a formatted DataFrame.")
    df = da.to_dataframe()
    df.reset_index(inplace=True)

    if not variable_mapper:
        df[UNIT_NAME] = da.units
        df[VARIABLE_NAME] = da.name.value
    else:
        # TODO: Make sure the mapping values are valid?
        df[UNIT_NAME] = variable_mapper[UNIT_NAME].get(da.units, da.units)
        df[VARIABLE_NAME] = variable_mapper[VARIABLE_NAME].get(
            da.name, da.name
        )
    df[CONFIGURATION_NAME] = f"{nwm_version}_retrospective"
    df[REFERENCE_TIME] = np.nan
    df[MEMBER] = None
    df.rename(
        columns={
            "time": VALUE_TIME,
            "feature_id": LOCATION_ID,
            da.name: VALUE,
        },
        inplace=True,
    )
    df.drop(columns=["latitude", "longitude"], inplace=True)

    df[LOCATION_ID] = f"{nwm_version}-" + df[LOCATION_ID].astype(str)

    if (nwm_version == "nwm21") or (nwm_version == "nwm30"):
        df.drop(columns=["elevation", "gage_id", "order"], inplace=True)

    return df


def datetime_to_date(dt: datetime) -> datetime:
    """Convert datetime to date only."""
    dt.replace(hour=0, minute=0, second=0, microsecond=0)
    return dt


def format_grouped_filename(ds_i: xr.Dataset) -> str:
    """Format the output filename based on min and max datetime."""
    min_year = ds_i.time.min().dt.year
    min_month = ds_i.time.min().dt.month
    min_day = ds_i.time.min().dt.day

    max_year = ds_i.time.max().dt.year
    max_month = ds_i.time.max().dt.month
    max_day = ds_i.time.max().dt.day

    min_time = f"{min_year.values}{min_month.values:02d}{min_day.values:02d}"
    max_time = f"{max_year.values}{max_month.values:02d}{max_day.values:02d}"

    if min_time == max_time:
        return f"{min_time}.parquet"
    else:
        return f"{min_time}_{max_time}.parquet"


@validate_call(config=dict(arbitrary_types_allowed=True))
def nwm_retro_to_parquet(
    nwm_version: SupportedNWMRetroVersionsEnum,
    variable_name: ChannelRtRetroVariableEnum,
    location_ids: List[int],
    start_date: Union[str, datetime, pd.Timestamp],
    end_date: Union[str, datetime, pd.Timestamp],
    output_parquet_dir: Union[str, Path],
    chunk_by: Union[NWMChunkByEnum, None] = None,
    overwrite_output: Optional[bool] = False,
    domain: Optional[SupportedNWMRetroDomainsEnum] = "CONUS",
    variable_mapper: Dict[str, Dict[str, str]] = None,
    timeseries_type: TimeseriesTypeEnum = "secondary"
):
    """Fetch NWM retrospective at NWM COMIDs and store as Parquet file.

    All dates and times within the files and in the file names are in UTC.

    Parameters
    ----------
    nwm_version : SupportedNWMRetroVersionsEnum
        NWM retrospective version to fetch.
        Currently `nwm20`, `nwm21`, and `nwm30` supported.
    variable_name : str
        Name of the NWM data variable to download.
        (e.g., "streamflow", "velocity", ...).
    location_ids : Iterable[int],
        NWM feature_ids to fetch.
    start_date : Union[str, datetime, pd.Timestamp]
        Date to begin data ingest.
        Str formats can include YYYY-MM-DD or MM/DD/YYYY
        Rounds down to beginning of day.

        - v2.0: 1993-01-01
        - v2.1: 1979-01-01
        - v3.0: 1979-02-01
    end_date : Union[str, datetime, pd.Timestamp],
        Last date to fetch.  Rounds up to end of day.
        Str formats can include YYYY-MM-DD or MM/DD/YYYY.

        - v2.0: 2018-12-31
        - v2.1: 2020-12-31
        - v3.0: 2023-01-31
    output_parquet_dir : Union[str, Path],
        Directory where output will be saved.
    chunk_by : Union[NWMChunkByEnum, None] = None,
        If None (default) saves all timeseries to a single file, otherwise
        the data is processed using the specified parameter.
        Can be: 'week', 'month', or 'year'.
    overwrite_output : bool = False,
        Whether output should overwrite files if they exist.  Default is False.
    domain : str = "CONUS"
        Geographical domain when NWM version is v3.0.
        Acceptable values are "Alaska", "CONUS" (default), "Hawaii", and "PR".
        Only used when NWM version equals `nwm30`.
    variable_mapper : Dict[str, Dict[str, str]]
        Dictionary mapping NWM variable and unit names to TEEHR variable
        and unit names.


    Examples
    --------
    Here we fetch and format retrospective NWM v2.0 streamflow data
    for two locations.

    Import the module.

    >>> import teehr.fetching.nwm.retrospective_points as nwm_retro

    Specify the input variables.

    >>> NWM_VERSION = "nwm20"
    >>> VARIABLE_NAME = "streamflow"
    >>> START_DATE = datetime(2000, 1, 1)
    >>> END_DATE = datetime(2000, 1, 2, 23)
    >>> LOCATION_IDS = [7086109, 7040481]
    >>> OUTPUT_ROOT = Path(Path().home(), "temp")
    >>> OUTPUT_DIR = Path(OUTPUT_ROOT, "nwm20_retrospective")

    Fetch and format the data, writing to the specified directory.

    >>> nwm_retro.nwm_retro_to_parquet(
    >>>     nwm_version=NWM_VERSION,
    >>>     variable_name=VARIABLE_NAME,
    >>>     start_date=START_DATE,
    >>>     end_date=END_DATE,
    >>>     location_ids=LOCATION_IDS,
    >>>     output_parquet_dir=OUTPUT_DIR
    >>> )
    """
    logger.info(
        f"Fetching NWM retrospective point data, version: {nwm_version}."
    )

    if nwm_version == SupportedNWMRetroVersionsEnum.nwm20:
        s3_zarr_url = "s3://noaa-nwm-retro-v2-zarr-pds"
    elif nwm_version == SupportedNWMRetroVersionsEnum.nwm21:
        s3_zarr_url = "s3://noaa-nwm-retrospective-2-1-zarr-pds/chrtout.zarr/"
    elif nwm_version == SupportedNWMRetroVersionsEnum.nwm30:
        s3_zarr_url = (
            f"s3://noaa-nwm-retrospective-3-0-pds/{domain}/zarr/chrtout.zarr"
        )
    else:
        raise ValueError(f"unsupported NWM version {nwm_version}")

    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)

    validate_retrospective_start_end_date(nwm_version, start_date, end_date)

    # Include the entirety of the specified end day
    end_date = end_date.to_period(freq="D").end_time

    output_dir = Path(output_parquet_dir)
    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    da = xr.open_zarr(
        fsspec.get_mapper(s3_zarr_url, anon=True), consolidated=True
    )[variable_name].sel(
        feature_id=location_ids, time=slice(start_date, end_date)
    )

    # Chunk data by time
    logger.debug("Chunking data by time.")
    periods = create_periods_based_on_chunksize(
        start_date=start_date,
        end_date=end_date,
        chunk_by=chunk_by
    )

    for period in periods:

        if period is not None:
            dts = get_period_start_end_times(
                period=period,
                start_date=start_date,
                end_date=end_date
            )
        else:
            dts = {"start_dt": start_date, "end_dt": end_date}

        da_i = da.sel(time=slice(dts["start_dt"], dts["end_dt"]))

        logger.debug(
            f"Fetching point data for {dts['start_dt']} to {dts['end_dt']}."
        )

        df = da_to_df(nwm_version, da_i, variable_mapper)
        output_filename = format_grouped_filename(da_i)
        output_filepath = Path(
            output_parquet_dir, output_filename
        )
        write_timeseries_parquet_file(
            filepath=output_filepath,
            overwrite_output=overwrite_output,
            data=df,
            timeseries_type=timeseries_type
        )



"""This module contains constants
 used throughout the package."""


# Primary evaluation directories
DATASET_DIR = "dataset"
CACHE_DIR = "cache"
SCRIPTS_DIR = "scripts"
LOCATIONS_DIR = "locations"
PRIMARY_TIMESERIES_DIR = "primary_timeseries"
LOCATION_CROSSWALKS_DIR = "location_crosswalks"
LOCATION_ATTRIBUTES_DIR = "location_attributes"
SECONDARY_TIMESERIES_DIR = "secondary_timeseries"
JOINED_TIMESERIES_DIR = "joined_timeseries"

# Fetching cache directories
FETCHING_CACHE_DIR = "fetching"
USGS_CACHE_DIR = "usgs"
NWM_CACHE_DIR = "nwm"
KERCHUNK_DIR = "kerchunk"
WEIGHTS_DIR = "weights"

# Loading cache directories
LOADING_CACHE_DIR = "loading"

S3_EVALUATIONS_PATH = "s3://ciroh-rti-public-data/teehr-data-warehouse/v0_4_evaluations/evaluations.yaml"



"""Module defining common utilities for fetching and processing NWM data."""
from pathlib import Path
from typing import Union, Optional, Iterable, List, Dict
from datetime import datetime
from datetime import timedelta
from dateutil.parser import parse
import logging
import re
import json

import dask
import fsspec
import ujson  # fast json
from kerchunk.hdf import SingleHdf5ToZarr
import pandas as pd
import numpy as np
import xarray as xr
import geopandas as gpd
import pyarrow as pa
import pandera

from teehr.models.fetching.utils import (
    SupportedKerchunkMethod,
    TimeseriesTypeEnum
)
from teehr.models.fetching.utils import (
    SupportedNWMOperationalVersionsEnum,
    NWMChunkByEnum
)
from teehr.fetching.const import (
    NWM_BUCKET,
    NWM_S3_JSON_PATH,
    NWM30_START_DATE,
    NWM21_START_DATE,
    NWM20_START_DATE,
    NWM12_START_DATE,
    NWM_VARIABLE_MAPPER,
    VARIABLE_NAME
)
import teehr.models.pandera_dataframe_schemas as schemas

TZ_PATTERN = re.compile(r't[0-9]+z')
DAY_PATTERN = re.compile(r'nwm.[0-9]+')


logger = logging.getLogger(__name__)


def start_on_z_hour(
    start_date: datetime,
    start_z_hour: int,
    gcs_component_paths: List[str]
):
    """Limit the start date to a specified z-hour."""
    logger.info(f"Limiting the start date to z-hour: {start_z_hour}.")
    formatted_start_date = start_date.strftime("%Y%m%d")
    return_list = []
    for path in gcs_component_paths:
        res = re.search(DAY_PATTERN, path).group()
        day = res.split(".")[1]
        tz = re.search(TZ_PATTERN, path).group()
        if day == formatted_start_date:
            if int(tz[1:-1]) >= start_z_hour:
                return_list.append(path)
        else:
            return_list.append(path)
    return return_list


def end_on_z_hour(
    start_date: datetime,
    ingest_days: int,
    end_z_hour: int,
    gcs_component_paths: List[str]
):
    """Limit the end date to a specified z-hour."""
    logger.info(f"Limiting the end date to z-hour: {end_z_hour}.")
    dates = pd.date_range(start=start_date, periods=ingest_days, freq="1d")
    formatted_end_date = dates[-1].strftime("%Y%m%d")
    return_list = []
    reversed_list = sorted(gcs_component_paths, reverse=True)
    for path in reversed_list:
        res = re.search(DAY_PATTERN, path).group()
        day = res.split(".")[1]
        tz = re.search(TZ_PATTERN, path).group()
        if day == formatted_end_date:
            if int(tz[1:-1]) <= end_z_hour:
                return_list.append(path)
        else:
            return_list.append(path)
    return sorted(return_list)


def parse_nwm_json_paths(
    json_paths: List[str]
) -> pd.DataFrame:
    """Parse the day and z-hour from the json paths, returning a DataFrame."""
    logger.debug("Parsing day and z-hour from json paths.")
    days = []
    z_hours = []
    for path in json_paths:
        filename = Path(path).name
        if path.split(":")[0] == "s3":
            res = re.search(DAY_PATTERN, path).group()
            days.append(res.split(".")[1])
            z_hours.append(re.search(TZ_PATTERN, filename).group())
        else:
            days.append(filename.split(".")[1])
            z_hours.append(filename.split(".")[3])

    return pd.DataFrame(
        {"day": days, "z_hour": z_hours, "filepath": json_paths}
    )


def format_nwm_configuration_name(
    nwm_configuration_name: str,
    nwm_version: str
) -> Dict[str, str]:
    """Format the NWM configuration name and member for the Evaluation.

    Returns a dictionary with the formatted configuration name and member,
    which is parsed from the NWM configuration name if it's an ensemble
    (ie., medium range or long range streamflow).
    """
    logger.info(
        f"Formatting configuration name for {nwm_configuration_name}."
    )
    ev_member = None
    if bool(re.search(r"_mem[0-9]+", nwm_configuration_name)):
        ev_configuration, ev_member = nwm_configuration_name.split("_mem")
    else:
        ev_configuration = nwm_configuration_name
    ev_configuration = nwm_version + "_" + ev_configuration
    return {"configuration_name": ev_configuration, "member": ev_member}


def format_nwm_variable_name(variable_name: str) -> str:
    """Format the NWM variable name for the Evaluation."""
    logger.info(f"Getting schema variable name for {variable_name}.")
    return NWM_VARIABLE_MAPPER[VARIABLE_NAME]. \
        get(variable_name, variable_name)


def validate_operational_start_end_date(
    nwm_version: str,
    start_date: Union[str, datetime],
    ingest_days: int
):
    """Make sure start/end dates work with specified NWM version."""
    logger.debug("Checking dates against NWM version.")

    if isinstance(start_date, str):
        start_date = parse(start_date)
    end_date = start_date + timedelta(days=ingest_days)

    err_msg = (
        f"The specified start and end dates ({start_date} - {end_date}) "
        f"fall outside {nwm_version} operational data availability."
    )
    v3_err_msg = (
        f"The specified start date ({start_date}) is before the NWM "
        f"v3.0 release date ({NWM30_START_DATE})"
    )

    if nwm_version == SupportedNWMOperationalVersionsEnum.nwm30:
        if start_date < NWM30_START_DATE:
            raise ValueError(v3_err_msg)
    if nwm_version == SupportedNWMOperationalVersionsEnum.nwm22:
        if (end_date >= NWM30_START_DATE) | (start_date < NWM21_START_DATE):
            raise ValueError(err_msg)
    if nwm_version == SupportedNWMOperationalVersionsEnum.nwm21:
        if (end_date >= NWM30_START_DATE) | (start_date < NWM21_START_DATE):
            raise ValueError(err_msg)
    if nwm_version == SupportedNWMOperationalVersionsEnum.nwm20:
        if (end_date >= NWM21_START_DATE) | (start_date < NWM20_START_DATE):
            raise ValueError(err_msg)
    if nwm_version == SupportedNWMOperationalVersionsEnum.nwm12:
        if (end_date >= NWM20_START_DATE) | (start_date < NWM12_START_DATE):
            raise ValueError(err_msg)


def generate_json_paths(
    kerchunk_method: str,
    gcs_component_paths: List[str],
    json_dir: str,
    ignore_missing_file: bool
) -> List[str]:
    """Generate file paths to Kerchunk reference json files.

    Parameters
    ----------
    kerchunk_method : str
        Specifies the preference in creating Kerchunk reference json files.
    gcs_component_paths : List[str]
        Paths to NWM netcdf files in GCS.
    json_dir : str
        Local directory for caching created json files.
    ignore_missing_file : bool
        Flag specifying whether or not to fail if a missing
        NWM file is encountered.

    Returns
    -------
    List[str]
        List of filepaths to json files locally and/or in s3.
    """
    logger.debug(f"Generating json paths. kerchunk_method: {kerchunk_method}")

    if kerchunk_method == SupportedKerchunkMethod.local:
        # Create them manually first
        json_paths = build_zarr_references(gcs_component_paths,
                                           json_dir,
                                           ignore_missing_file)

    elif kerchunk_method == SupportedKerchunkMethod.remote:
        # Use whatever pre-builts exist, skipping the rest
        fs = fsspec.filesystem("s3", anon=True)
        results = []
        for gcs_path in gcs_component_paths:
            results.append(check_for_prebuilt_json_paths(fs, gcs_path))
        json_paths = dask.compute(results)[0]
        json_paths = [path for path in json_paths if path is not None]

    elif kerchunk_method == SupportedKerchunkMethod.auto:
        # Use whatever pre-builts exist, and create the missing
        #  files, if any
        fs = fsspec.filesystem("s3", anon=True)
        results = []
        for gcs_path in gcs_component_paths:
            results.append(
                check_for_prebuilt_json_paths(
                    fs, gcs_path, return_gcs_path=True
                )
            )
        s3_or_gcs_paths = dask.compute(results)[0]

        # Build any jsons that do not already exist in s3
        gcs_paths = []
        json_paths = []
        for path in s3_or_gcs_paths:
            if path.split("://")[0] == "gcs":
                gcs_paths.append(path)
            else:
                json_paths.append(path)

        if len(gcs_paths) > 0:
            json_paths.extend(
                build_zarr_references(gcs_paths,
                                      json_dir,
                                      ignore_missing_file)
            )

    return json_paths


def _drop_nan_values(
    df: pd.DataFrame,
    subset_columns=["value"]
) -> pd.DataFrame:
    """Drop NaN values from the timeseries dataframe."""
    if df[subset_columns].isnull().values.any():
        logger.debug(
            "NaN values were encountered, dropping from the dataframe."
        )
        return df.dropna(subset=subset_columns).reset_index(drop=True)
    return df


def write_timeseries_parquet_file(
    filepath: Path,
    overwrite_output: bool,
    data: Union[pa.Table, pd.DataFrame],
    timeseries_type: TimeseriesTypeEnum
):
    """Write the output timeseries parquet file.

    Includes logic controlling whether or not to overwrite an existing file.

    Parameters
    ----------
    filepath : Path
        Path to the output parquet file.
    overwrite_output : bool
        Flag controlling overwrite behavior.
    data : Union[pa.Table, pd.DataFrame]
        The output data as either a dataframe or pyarrow table.
    """
    logger.debug(f"Writing parquet file: {filepath}")

    if isinstance(data, pa.Table):
        df = data.to_pandas()
    else:
        df = data

    df = _drop_nan_values(df)

    if timeseries_type == TimeseriesTypeEnum.primary:
        schema = schemas.primary_timeseries_schema(type="pandas")
    elif timeseries_type == TimeseriesTypeEnum.secondary:
        schema = schemas.secondary_timeseries_schema(type="pandas")

    try:
        validated_df = schema.validate(df, lazy=True)
    except pandera.errors.SchemaErrors as exc:
        msg = json.dumps(exc.message, indent=2)
        logger.error(
            f"Validation error: {msg}"
            f"\nThis file '{filepath}' will be skipped."
        )
        return

    if not filepath.is_file():
        validated_df.to_parquet(filepath)
    elif filepath.is_file() and overwrite_output:
        logger.info(f"Overwriting {filepath.name}")
        validated_df.to_parquet(filepath)
    elif filepath.is_file() and not overwrite_output:
        logger.info(
            f"{filepath.name} already exists and overwrite_output=False;"
            " skipping"
        )


def load_gdf(filepath: Union[str, Path], **kwargs: str) -> gpd.GeoDataFrame:
    """Load any supported geospatial file type into a gdf using GeoPandas."""
    logger.debug(f"Loading geospatial file: {filepath}")

    try:
        gdf = gpd.read_file(filepath, **kwargs)
        return gdf
    except Exception:
        pass
    try:
        gdf = gpd.read_parquet(filepath, **kwargs)
        return gdf
    except Exception:
        pass
    try:
        gdf = gpd.read_feather(filepath, **kwargs)
        return gdf
    except Exception:
        raise Exception("Unsupported zone polygon file type")


def parquet_to_gdf(parquet_filepath: str) -> gpd.GeoDataFrame:
    """Read parquet as GeoDataFrame."""
    gdf = gpd.read_parquet(parquet_filepath)
    return gdf


def np_to_list(t):
    """Convert numpy array to list."""
    return [a.tolist() for a in t]


def get_dataset(
    filepath: str, ignore_missing_file: bool, **kwargs
) -> xr.Dataset:
    """Retrieve a blob from the data service as xarray.Dataset.

    Parameters
    ----------
    filepath : str
        Path to the kerchunk json file. Can be local or remote.
    ignore_missing_file : bool
        Flag controlling whether to ignore missing files.

    Returns
    -------
    xarray.Dataset
        The data stored in the blob.
    """
    logger.debug(f"Getting xarray dataset from: {filepath}")

    try:
        m = fsspec.filesystem(
            "reference", fo=filepath, **kwargs
        ).get_mapper()
    except FileNotFoundError as e:
        if not ignore_missing_file:
            raise e
        else:
            return None
    except ValueError:
        raise ValueError(f"There was a problem reading {filepath}")
    return xr.open_dataset(m, engine="zarr", consolidated=False)


def list_to_np(lst):
    """Convert list to a tuple."""
    return tuple([np.array(a) for a in lst])


@dask.delayed
def check_for_prebuilt_json_paths(
    fs: fsspec.filesystem, gcs_path: str, return_gcs_path=False
) -> str:
    """Check for existence of a pre-built kerchunk json in s3 based \
    on its GCS path.

    Parameters
    ----------
    fs : fsspec.filesystem
        S3-based filesystem.
    gcs_path : str
        Path to the netcdf file in GCS.
    return_gcs_path : bool, optional
        Flag to return GCS path of s3 is missing, by default False.

    Returns
    -------
    str
        Path to the json in s3 or netcdf file in GCS.
    """
    s3_path = f"{NWM_S3_JSON_PATH}/{gcs_path.split('://')[1]}.json"
    if fs.exists(s3_path):
        return s3_path
    else:
        if return_gcs_path:
            return gcs_path


@dask.delayed
def gen_json(
    remote_path: str,
    fs: fsspec.filesystem,
    json_dir: Union[str, Path],
    ignore_missing_file: bool,
) -> str:
    """Create a single kerchunk reference JSON file.

    Parameters
    ----------
    remote_path : str
        Path to the file in the remote location (ie, GCS bucket).
    fs : fsspec.filesystem
        Fsspec filesystem mapped to GCS.
    json_dir : str
        Directory for saving zarr reference json files.

    Returns
    -------
    str
        Path to the local zarr reference json file.
    """
    so = dict(
        mode="rb",
        anon=True,
        default_fill_cache=False,
        default_cache_type="first",  # noqa
    )
    try:
        with fs.open(remote_path, **so) as infile:
            p = remote_path.split("/")
            date = p[3]
            fname = p[5]
            outf = str(Path(json_dir, f"{date}.{fname}.json"))
            try:
                h5chunks = SingleHdf5ToZarr(infile,
                                            remote_path,
                                            inline_threshold=300)
            except OSError as err:
                if not ignore_missing_file:
                    raise Exception(f"Corrupt file: {remote_path}") from err
                else:
                    logger.warning(
                        "A potentially corrupt file was encountered:"
                        f"{remote_path}"
                    )
                    return None
            with open(outf, "wb") as f:
                f.write(ujson.dumps(h5chunks.translate()).encode())
    except FileNotFoundError as e:
        if not ignore_missing_file:
            raise e
        else:
            logger.warning(f"A missing file was encountered: {remote_path}")
            return None
    return outf


def build_zarr_references(
    remote_paths: List[str],
    json_dir: Union[str, Path],
    ignore_missing_file: bool,
) -> list[str]:
    """Build the single file zarr json reference files using kerchunk.

    Parameters
    ----------
    remote_paths : List[str]
        List of remote filepaths.
    json_dir : str or Path
        Local directory for caching json files.

    Returns
    -------
    list[str]
        List of paths to the zarr reference json files.
    """
    logger.debug("Building zarr references.")

    json_dir_path = Path(json_dir)
    if not json_dir_path.exists():
        json_dir_path.mkdir(parents=True)

    fs = fsspec.filesystem("gcs", anon=True)

    # Check to see if the jsons already exist locally
    existing_jsons = []
    missing_paths = []
    for path in remote_paths:
        p = path.split("/")
        date = p[3]
        fname = p[5]
        local_path = Path(json_dir, f"{date}.{fname}.json")
        if local_path.exists():
            existing_jsons.append(str(local_path))
        else:
            missing_paths.append(path)
    if len(missing_paths) == 0:
        return sorted(existing_jsons)

    results = []
    for path in missing_paths:
        results.append(gen_json(path, fs, json_dir, ignore_missing_file))
    json_paths = dask.compute(results)[0]
    json_paths.extend(existing_jsons)

    if not any(json_paths):
        raise FileNotFoundError(
            "No NWM files for specified input configuration were found in GCS!"
        )

    json_paths = [path for path in json_paths if path is not None]

    return sorted(json_paths)


def construct_assim_paths(
    gcs_dir: str,
    configuration: str,
    output_type: str,
    dates: pd.DatetimeIndex,
    t_minus: Iterable[int],
    configuration_name_in_filepath: str,
    cycle_z_hours: Iterable[int],
    domain: str,
    file_extension: str = "nc"
) -> list[str]:
    """Construct paths to NWM point assimilation data.

    This function prioritizes value time over reference time so that only
    files with value times falling within the specified date range are included
    in the resulting file list.

    Parameters
    ----------
    gcs_dir : str
        Path to the NWM data on GCS.
    configuration : str
        Configuration type.
    output_type : str
        Output component of the configuration.
    dates : pd.DatetimeIndex
        Range of days to fetch data.
    t_minus : Iterable[int]
        Collection of lookback hours to include when fetching assimilation
        data.
    configuration_name_in_filepath : str
        Name of the assimilation configuration as represented in the GCS file.
        Defined in const_nwm.py.
    cycle_z_hours : Iterable[int]
        The z-hour of the assimilation configuration per day.
        Defined in const_nwm.py.
    domain : str
        Geographic region covered by the assimilation configuration.
        Defined in const_nwm.py.
    file_extension : str
        File extension ("nc" or "nc.json" for remote kerchunk).

    Returns
    -------
    list[str]
        List of remote filepaths.
    """
    logger.debug("Constructing assimilation paths.")

    component_paths = []

    for dt in dates:
        dt_str = dt.strftime("%Y%m%d")

        # Add the values starting from day 1,
        # skipping value times in the previous day
        if "hawaii" in configuration:
            for cycle_hr in cycle_z_hours:
                for tm in t_minus:
                    for tm2 in [0, 15, 30, 45]:
                        if (tm * 100 + tm2) > cycle_hr * 100:
                            continue
                        file_path = f"{gcs_dir}/nwm.{dt_str}/{configuration}/nwm.t{cycle_hr:02d}z.{configuration_name_in_filepath}.{output_type}.tm{tm:02d}{tm2:02d}.{domain}.{file_extension}"  # noqa
                        component_paths.append(file_path)
        else:
            for cycle_hr in cycle_z_hours:
                for tm in t_minus:
                    if tm > cycle_hr:
                        continue
                    file_path = f"{gcs_dir}/nwm.{dt_str}/{configuration}/nwm.t{cycle_hr:02d}z.{configuration_name_in_filepath}.{output_type}.tm{tm:02d}.{domain}.{file_extension}"  # noqa
                    component_paths.append(file_path)

        # Now add the values from the day following the end day,
        # whose value times that fall within the end day
        if "extend" in configuration:
            for tm in t_minus:
                dt_add = dt + pd.Timedelta(cycle_hr + 24, unit="hours")
                hr_add = dt_add.hour
                if tm > hr_add:
                    dt_add_str = dt_add.strftime("%Y%m%d")
                    file_path = f"{gcs_dir}/nwm.{dt_add_str}/{configuration}/nwm.t{hr_add:02d}z.{configuration_name_in_filepath}.{output_type}.tm{tm:02d}.{domain}.{file_extension}"  # noqa
                    component_paths.append(file_path)

        elif "hawaii" in configuration:
            for cycle_hr2 in cycle_z_hours:
                for tm in t_minus:
                    for tm2 in [0, 15, 30, 45]:
                        if cycle_hr2 > 0:
                            dt_add = dt + pd.Timedelta(
                                cycle_hr + cycle_hr2, unit="hours"
                            )
                            hr_add = dt_add.hour
                            if (tm * 100 + tm2) > hr_add * 100:
                                dt_add_str = dt_add.strftime("%Y%m%d")
                                file_path = f"{gcs_dir}/nwm.{dt_add_str}/{configuration}/nwm.t{hr_add:02d}z.{configuration_name_in_filepath}.{output_type}.tm{tm:02d}{tm2:02d}.{domain}.{file_extension}"  # noqa
                                component_paths.append(file_path)
        else:
            for cycle_hr2 in cycle_z_hours:
                for tm in t_minus:
                    if cycle_hr2 > 0:
                        dt_add = dt + pd.Timedelta(
                            cycle_hr + cycle_hr2, unit="hours"
                        )
                        hr_add = dt_add.hour
                        if tm > hr_add:
                            dt_add_str = dt_add.strftime("%Y%m%d")
                            file_path = f"{gcs_dir}/nwm.{dt_add_str}/{configuration}/nwm.t{hr_add:02d}z.{configuration_name_in_filepath}.{output_type}.tm{tm:02d}.{domain}.{file_extension}"  # noqa
                            component_paths.append(file_path)

    return sorted(component_paths)


def build_remote_nwm_filelist(
    configuration: str,
    output_type: str,
    start_dt: Union[str, datetime],
    ingest_days: int,
    analysis_config_dict: Dict,
    t_minus_hours: Optional[Iterable[int]],
    ignore_missing_file: Optional[bool],
    prioritize_analysis_valid_time: Optional[bool]
) -> List[str]:
    """Assemble a list of remote NWM files based on user parameters.

    Parameters
    ----------
    configuration : str
        Configuration type.
    output_type : str
        Output component of the configuration.
    start_dt : str “YYYY-MM-DD” or datetime
        Date to begin data ingest.
    ingest_days : int
        Number of days to ingest data after start date.
    t_minus_hours : Iterable[int]
        Only necessary if assimilation data is requested.
        Collection of lookback hours to include when fetching
        assimilation data.
    ignore_missing_file : bool
        Flag specifying whether or not to fail if a missing
        NWM file is encountered
        True = skip and continue
        False = fail.
    prioritize_analysis_valid_time : Optional[bool]
        A boolean flag that determines the method of fetching analysis data.
        When False (default), all hours of the reference time are included in the
        output. When True, only the hours within t_minus_hours are included.

    Returns
    -------
    list
        List of remote filepaths (strings).
    """
    logger.debug("Building remote NWM file list from GCS.")

    gcs_dir = f"gcs://{NWM_BUCKET}"
    fs = fsspec.filesystem("gcs", anon=True)
    dates = pd.date_range(start=start_dt, periods=ingest_days, freq="1d")

    if "assim" in configuration and prioritize_analysis_valid_time:
        cycle_z_hours = analysis_config_dict[configuration]["cycle_z_hours"]
        domain = analysis_config_dict[configuration]["domain"]
        configuration_name_in_filepath = analysis_config_dict[configuration][
            "configuration_name_in_filepath"
        ]
        max_lookback = analysis_config_dict[configuration]["num_lookback_hrs"]

        if max(t_minus_hours) > max_lookback - 1:
            raise ValueError(
                f"The maximum specified t-minus hour exceeds the lookback "
                f"period for this configuration: {configuration}; max t-minus: "  # noqa
                f"{max(t_minus_hours)} hrs; "
                f"look-back period: {max_lookback} hrs"
            )

        component_paths = construct_assim_paths(
            gcs_dir,
            configuration,
            output_type,
            dates,
            t_minus_hours,
            configuration_name_in_filepath,
            cycle_z_hours,
            domain,
        )

    else:
        component_paths = []

        for dt in dates:
            dt_str = dt.strftime("%Y%m%d")
            file_path = (
                f"{gcs_dir}/nwm.{dt_str}/{configuration}/nwm.*.{output_type}*"
            )
            result = fs.glob(file_path)
            if (len(result) == 0) & (not ignore_missing_file):
                raise FileNotFoundError(f"No NWM files found in {file_path}")
            component_paths.extend(result)
        component_paths = sorted([f"gcs://{path}" for path in component_paths])

    return component_paths


def get_period_start_end_times(
    period: pd.Period,
    start_date: datetime,
    end_date: datetime
) -> Dict[str, datetime]:
    """Get the start and end times for a period.

    Adjusts for the start and end dates of the total data ingest.

    Parameters
    ----------
    period : pd.Period
        The current period.
    start_date : datetime
        The start date of the data ingest.
    end_date : datetime
        Then end date of the data ingest.

    Returns
    -------
    Dict[str, datetime]
        The start and end times for the period.
    """
    logger.debug("Getting period start and end times.")

    start_dt = period.start_time
    end_dt = period.end_time

    if start_date > period.start_time:
        start_dt = start_date

    if (end_date < period.end_time) & (period.freq.name != "D"):
        end_dt = end_date

    return {"start_dt": start_dt, "end_dt": end_dt}


def create_periods_based_on_chunksize(
    start_date: datetime,
    end_date: datetime,
    chunk_by: Union[NWMChunkByEnum, None]
) -> List[pd.Period]:
    """Create a list of periods of a given frequency, start, and end time.

    Parameters
    ----------
    start_date : datetime
        The start date.
    end_date : datetime
        The end date.
    chunk_by : Union[NWMChunkByEnum, None]
        The chunk size frequency.

    Returns
    -------
    List[pd.Period]
        A pandas period range.
    """
    logger.debug("Creating periods based on chunk_by.")

    if chunk_by is None:
        periods = [None]

    if chunk_by == "day":
        periods = pd.period_range(start=start_date, end=end_date, freq="D")

    if chunk_by == "week":
        periods = pd.period_range(start=start_date, end=end_date, freq="W")

    if chunk_by == "month":
        periods = pd.period_range(start=start_date, end=end_date, freq="M")

    if chunk_by == "year":
        periods = pd.period_range(start=start_date, end=end_date, freq="Y")

    if chunk_by == "location_id":
        raise ValueError(
            "A period range cannot be created based on location_id."
        )

    return periods


def split_dataframe(df: pd.DataFrame, chunk_size: int) -> List[pd.DataFrame]:
    """Split a dataframe into chunks of a specified size.

    Parameters
    ----------
    df : pd.DataFrame
        The input dataframe.
    chunk_size : int
        The size of the chunks.

    Returns
    -------
    List[pd.DataFrame]
        A list of dataframes.
    """
    chunks = []
    for i in range(0, df.shape[0], chunk_size):
        chunks.append(df.iloc[i:i + chunk_size])
    return chunks



"""Module for generating area-based weights for grid layer pixels."""
from typing import Union, Dict
from pathlib import Path
import warnings
import logging

import geopandas as gpd
from geopandas import GeoDataFrame
import numpy as np
import xarray as xr
from rasterio.transform import rowcol
import rasterio
import pandas as pd
import dask
import shapely

from teehr.fetching.nwm.grid_utils import update_location_id_prefix
from teehr.fetching.utils import load_gdf
from teehr.fetching.const import LOCATION_ID
import teehr.models.pandera_dataframe_schemas as schemas

logger = logging.getLogger(__name__)


@dask.delayed
def vectorize(data_array: xr.DataArray) -> gpd.GeoDataFrame:
    """
    Convert 2D xarray.DataArray into a geopandas.GeoDataFrame.

    Notes
    -----
    Heavily borrowed from GeoCube, see:
    https://github.com/corteva/geocube/blob/master/geocube/vector.py#L12
    """
    # nodata mask
    mask = None
    if np.isnan(data_array.rio.nodata):
        mask = ~data_array.isnull()
    elif data_array.rio.nodata is not None:
        mask = data_array != data_array.rio.nodata

    # Give all pixels a unique value
    data_array.values[:, :] = np.arange(0, data_array.values.size).reshape(
        data_array.shape
    )

    # vectorize generator
    vectorized_data = (
        (value, shapely.geometry.shape(polygon))
        for polygon, value in rasterio.features.shapes(
            data_array,
            transform=data_array.rio.transform(),
            mask=mask,
        )
    )
    gdf = gpd.GeoDataFrame(
        vectorized_data,
        columns=[data_array.name, "geometry"],
        crs=data_array.rio.crs,
    )
    xx, yy = np.meshgrid(data_array.x.values, data_array.y.values)
    gdf["x"] = xx.ravel()
    gdf["y"] = yy.ravel()

    return gdf


@dask.delayed
def overlay_zones(
    grid: gpd.GeoDataFrame, zones: gpd.GeoDataFrame
) -> gpd.GeoDataFrame:
    """Overlay zone polygons on vectorized grid cells."""
    with pd.option_context(
        "mode.chained_assignment", None
    ):  # to ignore setwithcopywarning
        grid.loc[:, "pixel_area"] = grid.geometry.area
        overlay_gdf = grid.overlay(zones, keep_geom_type=True)
        overlay_gdf.loc[:, "overlay_area"] = overlay_gdf.geometry.area
        overlay_gdf.loc[:, "weight"] = (
            overlay_gdf.overlay_area / overlay_gdf.pixel_area
        )
    return overlay_gdf


def vectorize_grid(
    src_da: xr.DataArray,
    nodata_val: float,
    crs_wkt: str,
    vectorize_chunk: float = 40,
) -> gpd.GeoDataFrame:
    """Vectorize pixels in the template array in chunks using dask.

    Notes
    -----
    Parameter vectorize_chunk determines how many pixels will
    be vectorized at one time
    (thousands of pixels)
    """
    src_da = src_da.persist()
    max_pixels = vectorize_chunk * 1000
    num_splits = np.ceil(src_da.values.size / max_pixels).astype(int)

    # Prepare each data array
    if num_splits > 0:
        da_list = np.array_split(src_da, num_splits)
        [da.rio.write_nodata(nodata_val, inplace=True) for da in da_list]
    else:
        src_da.rio.write_nodata(nodata_val, inplace=True)
        da_list = [src_da]

    results = []
    for da_subset in da_list:
        results.append(vectorize(da_subset))
    grid_gdf = pd.concat(dask.compute(results)[0])
    grid_gdf.crs = crs_wkt

    # Reindex to remove duplicates
    grid_gdf["index"] = np.arange(len(grid_gdf.index))
    grid_gdf.set_index("index", inplace=True)

    return grid_gdf


def calculate_weights(
    grid_gdf: gpd.GeoDataFrame,
    zone_gdf: gpd.GeoDataFrame,
    overlay_chunk: float = 250,
) -> gpd.GeoDataFrame:
    """Overlay vectorized pixels and zone polygons, and calculate weights.

    Notes
    -----
    Parameter overlay_chunk determines the size of the rectangular
    window that spatially subsets datasets for the operation
    (thousands of pixels).
    """
    # Make sure geometries are valid
    grid_gdf["geometry"] = grid_gdf.geometry.make_valid()
    zone_gdf["geometry"] = zone_gdf.geometry.make_valid()

    xmin, ymin, xmax, ymax = zone_gdf.total_bounds

    x_steps = np.arange(xmin, xmax, overlay_chunk * 1000)
    y_steps = np.arange(ymin, ymax, overlay_chunk * 1000)

    x_steps = np.append(x_steps, xmax)
    y_steps = np.append(y_steps, ymax)

    results = []
    for i in range(x_steps.size - 1):
        for j in range(y_steps.size - 1):
            xmin = x_steps[i]
            xmax = x_steps[i + 1]

            ymin = y_steps[j]
            ymax = y_steps[j + 1]

            zone = zone_gdf.cx[xmin:xmax, ymin:ymax]
            grid = grid_gdf.cx[xmin:xmax, ymin:ymax]

            if len(zone.index) == 0 or len(grid.index) == 0:
                continue
            results.append(overlay_zones(grid, zone))

    overlay_gdf = pd.concat(dask.compute(results)[0])

    return overlay_gdf


def generate_weights_file(
    zone_polygons: Union[Path, str, GeoDataFrame],
    template_dataset: Union[str, Path, xr.Dataset],
    variable_name: str,
    output_weights_filepath: Union[str, Path],
    crs_wkt: str,
    unique_zone_id: str,
    location_id_prefix: str = None,
    **read_args: Dict,
) -> None:
    """Generate a file of area weights for pixels intersecting zone polyons.

    Parameters
    ----------
    zone_polygon_filepath : Union[Path, str, GeoDataFrame]
        Path to the polygons geoparquet file or GeoDataFrame.
    template_dataset : Union[str, Path, xr.Dataset]
        Path to the grid dataset or an xarray Dataset to use as a template.
    variable_name : str
        Name of the variable within the dataset.
    output_weights_filepath : str
        Path to the resultant weights file.
    crs_wkt : str
        Coordinate system for given template gridded dataset as WKT string.
        The zone_polygons will be reprojected to this CRS.
    unique_zone_id : str
        Name of the field in the zone polygon file containing unique IDs.
    location_id_prefix : str
        Prefix to add to the location_id field.
    **read_args : dict, optional
        Keyword arguments to be passed to GeoPandas read_file().
        read_parquet(), and read_feather() methods.

    Examples
    --------
    Here we generate weights for grid pixels intersecting a given
    polygon(s). The algorithm accounts for the fraction of the pixel
    area that is within the polygon. We'll use the Nextgen divides/
    catchments as the polygons and a NWM v2.2 forcing file as the
    template grid.

    Import the necessary modules.

    >>> from teehr.utilities.generate_weights import generate_weights_file
    >>> from teehr.fetching.nwm.const import CONUS_NWM_WKT

    Define the input variables.

    >>> TEST_DIR = Path("tests", "data", "nwm22")
    >>> TEMP_DIR = Path("tests", "data", "temp")
    >>> TEMPLATE_FILEPATH = Path(TEST_DIR, "test_template_grid.nc")
    >>> ZONES_FILEPATH = Path(TEST_DIR, "test_ngen_divides.parquet")
    >>> WEIGHTS_FILEPATH = Path(TEST_DIR, "test_weights_results.parquet")

    Perform the calculation, writing to the output directory, or optionally
    returning the dataframe if no output path is specified.

    >>> df = generate_weights_file(
    >>>     zone_polygon_filepath=ZONES_FILEPATH,
    >>>     template_dataset=TEMPLATE_FILEPATH,
    >>>     variable_name="RAINRATE",
    >>>     crs_wkt=CONUS_NWM_WKT,
    >>>     output_weights_filepath=None,
    >>>     location_id_prefix="ngen",
    >>>     unique_zone_id="id",
    >>> )
    """
    if unique_zone_id is None:
        logger.error("unique_zone_id must be provided.")
        raise ValueError("unique_zone_id must be provided.")

    if isinstance(zone_polygons, (str, Path)):
        zone_gdf = load_gdf(zone_polygons, **read_args)
        zone_gdf = zone_gdf.to_crs(crs_wkt)
    elif isinstance(zone_polygons, GeoDataFrame):
        zone_gdf = zone_polygons.to_crs(crs_wkt)
    else:
        logger.error(
            "zone_polygons must be a path to a file or a GeoDataFrame."
        )
        raise ValueError(
            "zone_polygons must be a path to a file or a GeoDataFrame."
        )

    if isinstance(template_dataset, (str, Path)):
        template_dataset = xr.open_dataset(template_dataset)
    src_da = template_dataset[variable_name]
    src_da = src_da.rio.write_crs(crs_wkt, inplace=True)
    grid_transform = src_da.rio.transform()
    nodata_val = src_da.rio.nodata

    if not all([dim in src_da.dims for dim in ["x", "y"]]):
        raise ValueError("Template dataset must have x and y dimensions.")

    # Get the subset of the grid that intersects the total zone bounds
    bbox = tuple(zone_gdf.total_bounds)
    if len(template_dataset.dims) == 2:
        src_da = src_da.sel(
            x=slice(bbox[0], bbox[2]), y=slice(bbox[1], bbox[3])
        )
    else:
        src_da = src_da.sel(
            x=slice(bbox[0], bbox[2]), y=slice(bbox[1], bbox[3])
        )[0]
    src_da = src_da.astype("float32")
    src_da["x"] = np.float32(src_da.x.values)
    src_da["y"] = np.float32(src_da.y.values)

    # Vectorize source grid pixels
    grid_gdf = vectorize_grid(src_da, nodata_val, crs_wkt)

    # Overlay and calculate areal weights of pixels within each zone
    # Note: Temporarily suppress the dask UserWarning: "Large object detected
    #  in task graph" until a better approach is found
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        weights_gdf = calculate_weights(grid_gdf, zone_gdf)
    weights_gdf = weights_gdf.drop_duplicates(
        subset=["x", "y", unique_zone_id]
    )

    # Convert x-y to row-col using original transform
    rows, cols = rowcol(
        grid_transform, weights_gdf.x.values, weights_gdf.y.values
    )
    weights_gdf["row"] = rows
    weights_gdf["col"] = cols

    if unique_zone_id:
        df = weights_gdf[["row", "col", "weight", unique_zone_id]].copy()
        df.rename(columns={unique_zone_id: LOCATION_ID}, inplace=True)
    else:
        df = weights_gdf[["row", "col", "weight"]]
        df[LOCATION_ID] = weights_gdf.index.values

    if location_id_prefix:
        df = update_location_id_prefix(
            df, new_prefix=location_id_prefix
        )

    schema = schemas.weights_file_schema()
    validated_df = schema.validate(df)
    validated_df.to_parquet(output_weights_filepath)

    return



"""Module describing NWM v2.0 grid configuration variables."""
from teehr.models.str_enum import StrEnum

from typing import Optional

from pydantic import BaseModel

from teehr.models.fetching.nwm22_grid import (
    Analysis,
    ShortRange,
    MediumRange,
    LongRange,
    Forcing,
)


# CONFIGURATIONS ENUM
class ConfigurationsEnum(StrEnum):
    """ConfigurationsEnum."""

    analysis_assim = "analysis_assim"
    analysis_assim_extend = "analysis_assim_extend"
    analysis_assim_long = "analysis_assim_long"
    analysis_assim_hawaii = "analysis_assim_hawaii"
    short_range = "short_range"
    short_range_hawaii = "short_range_hawaii"
    medium_range_mem1 = "medium_range_mem1"
    medium_range_mem2 = "medium_range_mem2"
    medium_range_mem3 = "medium_range_mem3"
    medium_range_mem4 = "medium_range_mem4"
    medium_range_mem5 = "medium_range_mem5"
    medium_range_mem6 = "medium_range_mem6"
    medium_range_mem7 = "medium_range_mem7"
    long_range_mem1 = "long_range_mem1"
    long_range_mem2 = "long_range_mem2"
    long_range_mem3 = "long_range_mem3"
    long_range_mem4 = "long_range_mem4"
    forcing_medium_range = "forcing_medium_range"
    forcing_short_range = "forcing_short_range"
    forcing_short_range_hawaii = "forcing_short_range_hawaii"
    forcing_analysis_assim = "forcing_analysis_assim"
    forcing_analysis_assim_extend = "forcing_analysis_assim_extend"
    forcing_analysis_assim_hawaii = "forcing_analysis_assim_hawaii"


# CONFIGURATION MODEL
class GridConfigurationModel(BaseModel):
    """NWM v2.0 GridConfigurationModel."""

    configuration: ConfigurationsEnum
    analysis_assim: Optional[Analysis] = None
    analysis_assim_extend: Optional[Analysis] = None
    analysis_assim_long: Optional[Analysis] = None
    analysis_assim_hawaii: Optional[Analysis] = None
    short_range: Optional[ShortRange] = None
    short_range_hawaii: Optional[ShortRange] = None
    medium_range_mem1: Optional[MediumRange] = None
    medium_range_mem2: Optional[MediumRange] = None
    medium_range_mem3: Optional[MediumRange] = None
    medium_range_mem4: Optional[MediumRange] = None
    medium_range_mem5: Optional[MediumRange] = None
    medium_range_mem6: Optional[MediumRange] = None
    medium_range_mem7: Optional[MediumRange] = None
    long_range_mem1: Optional[LongRange] = None
    long_range_mem2: Optional[LongRange] = None
    long_range_mem3: Optional[LongRange] = None
    long_range_mem4: Optional[LongRange] = None
    forcing_medium_range: Optional[Forcing] = None
    forcing_short_range: Optional[Forcing] = None
    forcing_short_range_hawaii: Optional[Forcing] = None
    forcing_analysis_assim: Optional[Forcing] = None
    forcing_analysis_assim_extend: Optional[Forcing] = None
    forcing_analysis_assim_hawaii: Optional[Forcing] = None


if __name__ == "__main__":
    # So for example:
    configuration = "forcing_medium_range"
    output_type = "forcing"
    variable_name = "RAINRATE"

    # Assemble input parameters
    vars = {
        "configuration": configuration,
        configuration: {
            "output_type": output_type,
            output_type: variable_name,
        },
    }
    cm = GridConfigurationModel.model_validate(vars)

    config = cm.configuration.name
    forecast_obj = getattr(cm, config)
    out_type = forecast_obj.output_type.name
    var_name = getattr(forecast_obj, out_type).name



"""Initialize the TEEHR package."""
__version__ = "0.4.10"

from teehr.evaluation.evaluation import Evaluation  # noqa
from teehr.models.metrics.deterministic_models import DeterministicMetrics  # noqa
from teehr.models.metrics.probabilistic_models import ProbabilisticMetrics  # noqa
from teehr.models.metrics.signature_models import SignatureMetrics  # noqa
from teehr.models.metrics.bootstrap_models import Bootstrappers  # noqa
from teehr.models.metrics.basemodels import Operators  # noqa
from teehr.models.pydantic_table_models import (  # noqa
    Configuration,
    Attribute,
    Unit,
    Variable
)
from teehr.models.filters import (  # noqa
    UnitFilter,
    ConfigurationFilter,
    VariableFilter,
    AttributeFilter,
    LocationFilter,
    LocationAttributeFilter,
    LocationCrosswalkFilter,
    TimeseriesFilter,
    JoinedTimeseriesFilter
)

from teehr.models.calculated_fields.row_level import RowLevelCalculatedFields # noqa
from teehr.models.calculated_fields.timeseries_aware import TimeseriesAwareCalculatedFields # noqa


"""Utility functions for querying data."""
import geopandas as gpd
import pandas as pd
from typing import List, Union
from teehr.models.str_enum import StrEnum
import pyspark.sql as ps
from pyspark.sql import DataFrame
import pyspark.sql.functions as F

import logging

logger = logging.getLogger(__name__)


def unpack_sdf_dict_columns(sdf: DataFrame, column_name: str) -> DataFrame:
    """Explode a column of dictionaries into new columns named key_item."""
    first_row = sdf.select(column_name).first()
    keys = sdf.sparkSession.createDataFrame([first_row]).select(
        F.explode(F.map_keys(F.col(column_name)))
        ).distinct()
    key_list = list(map(lambda row: row[0], keys.collect()))
    key_cols = list(
        map(lambda f: F.col(column_name).getItem(f).alias(str(f)), key_list)
    )
    df_cols = list(sdf.columns)
    df_cols.remove(column_name)
    return sdf.select(*df_cols, *key_cols)


def df_to_gdf(df: pd.DataFrame) -> gpd.GeoDataFrame:
    """Convert pd.DataFrame to gpd.GeoDataFrame.

    When the `geometry` column is read from a parquet file using DuckBD
    it is a bytearray in the resulting pd.DataFrame.  The `geometry` needs
    to be convert to bytes before GeoPandas can work with it.  This function
    does that.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with a `geometry` column that has geometry stored as
        a bytearray.

    Returns
    -------
    gpd.GeoDataFrame
        GeoDataFrame with a valid `geometry` column.
    """
    df["geometry"] = gpd.GeoSeries.from_wkb(
        df["geometry"].apply(lambda x: bytes(x))
    )
    return gpd.GeoDataFrame(df, crs="EPSG:4326", geometry="geometry")


def validate_fields_exist(
        valid_fields: List[str],
        requested_fields: List[str]
):
    """Validate that the requested_fields are in the valid_fields list."""
    logger.debug("Validating requested fields.")
    if not all(e in valid_fields for e in requested_fields):
        error_msg = f"One of the requested fields: {requested_fields} is not" \
                    f" a valid DataFrame field: {valid_fields}."
        logger.error(error_msg)
        raise ValueError(error_msg)


def parse_fields_to_list(
        requested_fields: Union[str, StrEnum, List[Union[str, StrEnum]]]
) -> List[str]:
    """Convert the requested fields to a list of strings."""
    logger.debug("Parsing requested fields to a list of strings.")
    if not isinstance(requested_fields, List):
        requested_fields = [requested_fields]
    requested_fields_strings = []
    for field in requested_fields:
        if isinstance(field, str):
            requested_fields_strings.append(field)
        else:
            requested_fields_strings.append(field.value)
    return requested_fields_strings


def order_df(df, sort_by: Union[str, StrEnum, List[Union[str, StrEnum]]]):
    """Sort a DataFrame by a list of columns."""
    logger.debug("Ordering DataFrame.")
    sort_by_strings = parse_fields_to_list(sort_by)
    validate_fields_exist(df.columns, sort_by_strings)
    return df.orderBy(*sort_by_strings)


def group_df(df, group_by: Union[str, StrEnum, List[Union[str, StrEnum]]]):
    """Group a DataFrame by a list of columns."""
    logger.debug("Grouping DataFrame.")
    group_by_strings = parse_fields_to_list(group_by)
    validate_fields_exist(df.columns, group_by_strings)
    return df.groupBy(*group_by_strings)


def join_geometry(
    target_df: ps.DataFrame,
    location_df: ps.DataFrame,
    target_location_id: str = "location_id",
):
    """Join geometry."""
    logger.debug("Joining locations geometry.")

    joined_df = target_df.join(
        location_df.withColumnRenamed(
            "id",
            target_location_id
        ).select(
            target_location_id,
            "geometry"
        ), on=target_location_id
    )
    return df_to_gdf(joined_df.toPandas())



"""Utility functions for the evaluation class."""
import logging
import fnmatch
from typing import List
from pathlib import Path

logger = logging.getLogger(__name__)


PIPE = "│"
ELBOW = "└──"
TEE = "├──"
PIPE_PREFIX = "│   "
SPACE_PREFIX = "    "


def print_tree(
    path,
    prefix="",
    exclude_patterns: List[str] = [""],
    max_depth: int = -1,
    current_depth: int = 0
):
    """Print the directory tree structure."""
    if max_depth != -1 and current_depth > max_depth:
        return

    # Get all files and directories in the current path
    paths = list(Path(path).glob("*"))

    # Exclude specific files and directories
    filtered_files = [f for f in paths if not any(fnmatch.fnmatch(f.name, p) for p in exclude_patterns)]

    # Print the directory tree structure
    entries_count = len(filtered_files)

    for index, full_path in enumerate(filtered_files):
        name = full_path.relative_to(path)
        connector = ELBOW if index == entries_count - 1 else TEE

        if full_path.is_dir():
            print(f"{prefix}{connector} {name}")
            new_prefix = prefix + (PIPE_PREFIX if index != entries_count - 1 else SPACE_PREFIX)
            print_tree(
                path=full_path,
                prefix=new_prefix,
                exclude_patterns=exclude_patterns,
                max_depth=max_depth,
                current_depth=current_depth + 1
            )
        else:
            print(f"{prefix}{connector} {name}")



"""Functions for formatting filters for querying."""
import pandas as pd
import pandera as pa
import warnings

from collections.abc import Iterable
from datetime import datetime
from typing import List, Union
import logging
from pyspark.sql import DataFrame
from teehr.models.str_enum import StrEnum

from teehr.models.filters import FilterBaseModel
from teehr.models.pydantic_table_models import TableBaseModel

logger = logging.getLogger(__name__)

SQL_DATETIME_STR_FORMAT = "%Y-%m-%d %H:%M:%S"


def get_datetime_list_string(values):
    """Get a datetime list as a list of strings."""
    return [f"'{v.strftime(SQL_DATETIME_STR_FORMAT)}'" for v in values]


def format_iterable_value(
    values: Iterable[Union[str, int, float, datetime]]
) -> str:
    """Return an SQL formatted string from list of values.

    Parameters
    ----------
    values : Iterable
        Contains values to be formatted as a string for SQL. Only one type of
        value (str, int, float, datetime) should be used. First value in list
        is used to determine value type. Values are not checked for type
        consistency.

    Returns
    -------
    str
        An SQL formatted string from list of values.
    """
    # string
    if isinstance(values[0], str):
        return f"""({",".join([f"'{v}'" for v in values])})"""
    # int or float
    elif isinstance(values[0], int) or isinstance(values[0], float):
        return f"""({",".join([f"{v}" for v in values])})"""
    # datetime
    elif isinstance(values[0], datetime):
        return f"""({",".join(get_datetime_list_string(values))})"""
    else:
        logger.warning(
            "Treating value as string because didn't know what else to do."
        )
        return f"""({",".join([f"'{str(v)}'" for v in values])})"""


def format_filter(
    filter: FilterBaseModel,
) -> str:
    r"""Return an SQL formatted string for single filter object.

    Parameters
    ----------
    filter : str or FilterBaseModel
        A single FilterBaseModel object or a subclass
        of FilterBaseModel.

    Returns
    -------
    str
        An SQL formatted string for single filter object.
    """
    column = filter.column.value
    operator = filter.operator.value

    if isinstance(filter.value, str):
        return f"""{column} {operator} '{filter.value}'"""
    elif (
        isinstance(filter.value, int)
        or isinstance(filter.value, float)
    ):
        return f"""{column} {operator} {filter.value}"""
    elif isinstance(filter.value, datetime):
        dt_str = filter.value.strftime(SQL_DATETIME_STR_FORMAT)
        return f"""{column} {operator} '{dt_str}'"""
    elif (
        isinstance(filter.value, Iterable)
        and not isinstance(filter.value, str)
    ):
        value = format_iterable_value(filter.value)
        return f"""{column} {operator} {value}"""
    else:
        logger.warning(
            f"Treating value {filter.value} as string because "
            "didn't know what else to do."
        )
        return f"""{column} {operator} '{str(filter.value)}'"""


def validate_filter(
    filter: FilterBaseModel,
    dataframe_schema: pa.DataFrameSchema
):
    """Validate a single model."""
    if filter.column.value not in dataframe_schema.columns:
        raise ValueError(f"Filter column not in model fields: {filter}")

    model_field_data_type = dataframe_schema.columns[filter.column.value].dtype.type
    logging.debug(
        f"Model field {filter.column.value} has type: {model_field_data_type}"
    )

    # if string or not iterable, make it iterable
    vals = filter.value
    value_was_iterable = True
    if isinstance(vals, str) or not isinstance(vals, Iterable):
        vals = [vals]
        value_was_iterable = False

    validate_vals = []
    for v in vals:
        logging.debug(f"Validating filter value: {v}")
        if model_field_data_type == str:
            validate_vals.append(str(v))
        elif model_field_data_type == int:
            validate_vals.append(int(v))
        elif (
            model_field_data_type == float or
            model_field_data_type == "float64" or
            model_field_data_type == "float32"
        ):
            validate_vals.append(float(v))
        elif (
            model_field_data_type == datetime or
            model_field_data_type == "datetime64[ns]" or
            model_field_data_type == "datetime64[ms]"
        ):
            validate_vals.append(pd.Timestamp(v))
        else:
            logger.warning(
                f"Treating value as string because "
                "didn't know what else to do."
            )
            validate_vals.append(str(v))

    if value_was_iterable:
        filter.value = validate_vals
    else:
        filter.value = validate_vals[0]

    return filter


def validate_and_apply_filters(
    sdf: DataFrame,
    filters: Union[str, dict, List[dict]],
    filter_model: FilterBaseModel,
    fields_enum: StrEnum,
    dataframe_schema: pa.DataFrameSchema = None,
    validate: bool = True
):
    """Validate and apply filters."""
    if isinstance(filters, str):
        logger.debug(f"Filter {filters} is already string.  Applying as is.")
        sdf = sdf.filter(filters)
        return sdf

    if not isinstance(filters, List):
        logger.debug("Filter is not a list.  Making a list.")
        filters = [filters]

    for filter in filters:
        logger.debug(f"Validating and applying {filter}")

        if not isinstance(filter, str):
            filter = filter_model.model_validate(
                filter,
                context={"fields_enum": fields_enum}
            )
            logger.debug(f"Filter: {filter.model_dump_json()}")
            if validate:
                filter = validate_filter(filter, dataframe_schema)
            filter = format_filter(filter)

        sdf = sdf.filter(filter)

    return sdf



from teehr.evaluation.tables.domain_table import DomainTable
from teehr.models.filters import UnitFilter
from teehr.models.table_enums import UnitFields
from teehr.models.pydantic_table_models import Unit
import teehr.models.pandera_dataframe_schemas as schemas
from typing import List, Union
import logging
from teehr.utils.utils import to_path_or_s3path


logger = logging.getLogger(__name__)


class UnitTable(DomainTable):
    """Access methods to units table."""

    def __init__(self, ev):
        """Initialize class."""
        super().__init__(ev)
        self.name = "units"
        self.dir = to_path_or_s3path(ev.dataset_dir, self.name)
        self.filter_model = UnitFilter
        self.schema_func = schemas.unit_schema
        self.unique_column_set = ["name"]

    def field_enum(self) -> UnitFields:
        """Get the unit fields enum."""
        fields = self._get_schema("pandas").columns.keys()
        return UnitFields(
            "UnitFields",
            {field: field for field in fields}
        )

    def add(
        self,
        unit: Union[Unit, List[Unit]]
    ):
        """Add a unit to the evaluation.

        Parameters
        ----------
        unit : Union[Unit, List[Unit]]
            The unit domain to add.

        Example
        -------
        >>> from teehr.models.domain_tables import Unit
        >>> unit = Unit(
        >>>     name="m^3/s",
        >>>     long_name="Cubic meters per second"
        >>> )
        >>> ev.load.add_unit(unit)
        """
        self._add(unit)


"""Download and copy AWS jars to SPARK install."""
import pyspark
import urllib.request


def main():
    """Download and copy AWS jars to SPARK install."""
    SPARK_HOME = pyspark.__path__[0]
    print(f"SPARK_HOME is: {SPARK_HOME}")

    SPARK_VERSION = pyspark.__version__
    print(f"SPARK_VERSION is: {SPARK_VERSION}")

    jars = [
        {
            "url": "https://repo1.maven.org/maven2/com/amazonaws/aws-java-sdk-bundle/1.12.262/aws-java-sdk-bundle-1.12.262.jar", # noqa
            "path": f"{SPARK_HOME}/jars/aws-java-sdk-bundle-1.12.262.jar"
        },
        {
            "url": "https://repo1.maven.org/maven2/org/apache/hadoop/hadoop-aws/3.3.4/hadoop-aws-3.3.4.jar", # noqa
            "path": f"{SPARK_HOME}/jars/hadoop-aws-3.3.4.jar"
        }
    ]

    for jar in jars:
        print(f"Downloading {jar['url']} to {jar['path']}")
        urllib.request.urlretrieve(jar['url'], jar['path'])
        print(f"Downloaded {jar['url']} to {jar['path']}")


if __name__ == "__main__":
    main()


"""Classes representing UDFs."""
from typing import List, Union
from pydantic import Field
import pandas as pd
import pyspark.sql.types as T
import pyspark.sql as ps
from teehr.models.calculated_fields.base import CalculatedFieldABC, CalculatedFieldBaseModel

class PercentileEventDetection(CalculatedFieldABC, CalculatedFieldBaseModel):
    """Adds an "event" and "event_id" column to the DataFrame based on a percentile threshold.

    The "event" column (bool) indicates whether the value is above the XXth percentile.
    The "event_id" column (string) groups continuous segments of events and assigns a
    unique ID to each segment in the format "startdate-enddate".

    Properties
    ----------
    - quantile:
        The percentile threshold to use for event detection.
        Default: 0.85
    - value_time_field_name:
        The name of the column containing the timestamp.
        Default: "value_time"
    - value_field_name:
        The name of the column containing the value to detect events on.
        Default: "primary_value"
    - output_event_field_name:
        The name of the column to store the event detection.
        Default: "event"
    - output_event_id_field_name:
        The name of the column to store the event ID.
        Default: "event_id"
    - uniqueness_fields:
        The columns to use to uniquely identify each timeseries.

        .. code-block:: python

            Default: [
                'reference_time',
                'primary_location_id',
                'configuration_name',
                'variable_name',
                'unit_name'
            ]
    """
    quantile: float = Field(
        default=0.85
    )
    value_time_field_name: str = Field(
        default="value_time"
    )
    value_field_name: str = Field(
        default="primary_value"
    )
    output_event_field_name: str = Field(
        default="event"
    )
    output_event_id_field_name: str = Field(
        default="event_id"
    )
    uniqueness_fields: Union[str, List[str]] = Field(
        default=[
            'reference_time',
            'primary_location_id',
            'configuration_name',
            'variable_name',
            'unit_name'
        ]
    )

    @staticmethod
    def add_is_event(
        sdf: ps.DataFrame,
        output_field,
        input_field,
        quantile,
        group_by,
        return_type=T.BooleanType()

    ):
        # Get the schema of the input DataFrame
        input_schema = sdf.schema

        # Create a copy of the schema and add the new column
        output_schema = T.StructType(input_schema.fields + [T.StructField(output_field, return_type, True)])

        def is_event(pdf, input_field, quantile, output_field) -> pd.DataFrame:
            pvs = pdf[input_field]

            # Calculate the XXth percentile
            percentile = pvs.quantile(quantile)

            # Create a new column indicating whether each value is above the XXth percentile
            pdf[output_field] = pvs > percentile

            return pdf

        def wrapper(pdf, input_field, quantile, output_field):
            return is_event(pdf, input_field, quantile, output_field)

        # Group the data and apply the UDF
        # lambda pdf: wrapper_function(pdf, threshold_value)
        sdf = sdf.groupby(group_by).applyInPandas(
            lambda pdf: wrapper(pdf, input_field, quantile, output_field),
            schema=output_schema
        )

        return sdf

    @staticmethod
    def add_event_ids(
        sdf,
        output_field,
        input_field,
        time_field,
        group_by,
        return_type=T.StringType(),

    ):
        # Get the schema of the input DataFrame
        input_schema = sdf.schema

        # Create a copy of the schema and add the new column
        output_schema = T.StructType(input_schema.fields + [T.StructField(output_field, return_type, True)])

        def event_ids(pdf: pd.DataFrame, input_field, time_field, output_field) -> pd.DataFrame:
            # Create a new column for continuous segments
            pdf['segment'] = (pdf[input_field] != pdf[input_field].shift()).cumsum()

            # Filter only the segments where values are over the 90th percentile
            segments = pdf[pdf[input_field]]

            # Group by segment and create startdate-enddate string
            segment_ranges = segments.groupby('segment').agg(
                startdate=(time_field, 'min'),
                enddate=(time_field, 'max')
            ).reset_index()

            # Merge the segment ranges back to the original DataFrame
            pdf = pdf.merge(segment_ranges[['segment', 'startdate', 'enddate']], on='segment', how='left')

            # Create the startdate-enddate string column
            pdf[output_field] = pdf.apply(
                lambda row: f"{row['startdate']}-{row['enddate']}" if pd.notnull(row['startdate']) else None,
                axis=1
            )

            # Drop the 'segment', 'startdate', and 'enddate' columns before returning
            pdf.drop(columns=['segment', 'startdate', 'enddate'], inplace=True)

            return pdf

        def wrapper(pdf, input_field, time_field, output_field):
            return event_ids(pdf, input_field, time_field, output_field)

        # Group the data and apply the UDF
        sdf = sdf.orderBy(*group_by, time_field).groupby(group_by).applyInPandas(
            lambda pdf: wrapper(pdf, input_field, time_field, output_field),
            schema=output_schema
        )

        return sdf

    def apply_to(self, sdf: ps.DataFrame) -> ps.DataFrame:

        sdf = self.add_is_event(
            sdf=sdf,
            input_field=self.value_field_name,
            quantile=self.quantile,
            output_field=self.output_event_field_name,
            group_by=self.uniqueness_fields
        )
        sdf = self.add_event_ids(
            sdf=sdf,
            input_field=self.output_event_field_name,
            time_field=self.value_time_field_name,
            output_field=self.output_event_id_field_name,
            group_by=self.uniqueness_fields
        )

        return sdf


class TimeseriesAwareCalculatedFields():
    """Timeseries aware calculated fields.

    Notes
    -----
    Timeseries aware CFs are aware of ordered groups of data (e.g., a timeseries).
    This is useful for things such as event detection, base flow separation, and
    other fields that need to be calculated based on a entire timeseries.  The
    definition of what creates a unique set of timeseries (i.e., a timeseries) can
    be specified.

    Available Calculated Fields:

    - PercentileEventDetection
    """

    PercentileEventDetection = PercentileEventDetection


"""Bootstrap classes."""
from typing import Any, Callable, Union
from collections.abc import Generator as PyGenerator
from pathlib import Path
import logging

from arch.bootstrap import IIDBootstrap
from arch.bootstrap.base import (
    # _loo_jackknife, # leave one out jackknife
    _add_extra_kwargs
)
from arch.typing import ArrayLike, Int64Array, Float64Array
from numpy.random import Generator, RandomState
import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)

# FILL_VALUE = -9999  # JAB


class GumbootBootstrap(IIDBootstrap):
    """Python implementation of the Gumboot R package."""

    def __init__(
        self,
        *args: ArrayLike,
        value_time: pd.Series,
        water_year_month: Union[int, None],
        boot_year_file: Union[str, Path, None],
        seed: Union[int, Generator, RandomState, None],
        **kwargs: ArrayLike,
    ) -> None:
        """Initialize the Gumboot class, inheriting from IIDBootstrap."""
        super().__init__(*args, seed=seed, **kwargs)

        # JAB
        # self.boot_year_array: Union[np.array, None]

        logger.debug("Initializing the Gumboot water years.")

        value_time = pd.to_datetime(value_time)
        water_years = np.where(
            value_time.dt.month >= water_year_month,
            value_time.dt.year + 1,
            value_time.dt.year
        )  # (iyWater)

        self.unique_water_years = np.unique(water_years)
        self.num_water_years = self.unique_water_years.size
        self.water_year_array = water_years

        if boot_year_file:
            self.user_defined_boot_years = True
            self.boot_year_array = np.genfromtxt(
                boot_year_file, delimiter=",", dtype=int
            )
            if np.in1d(
                np.unique(self.boot_year_array),
                self.unique_water_years
            ).all() is False:
                logger.error(
                    "Invalid boot year file: The provided boot year file "
                    "contains years outside of the range of the timeseries "
                    "data."
                )
        else:
            self.user_defined_boot_years = False

    def update_indices(
        self,
        rep: int
    ) -> Int64Array:  # type: ignore
        """
        Gumboot resampling implementation.

        Notes
        -----
        This requires an extra field (value_time) to create indices based on
        water year (time).
        """
        logger.debug(f"Resampling the Gumboot water years. rep: {rep}")

        if rep == 0:
            jx_valid = np.arange(len(self._args[0]))
        else:
            if self.user_defined_boot_years:
                years = self.boot_year_array[:, rep - 1]
            else:
                year_indexes = self._generator.integers(
                    self.num_water_years, size=self.num_water_years
                )
                years = np.array(self.unique_water_years)[year_indexes]
                # self.boot_year_array[:, rep - 1] = years  # JAB

            # Get the indexes of values corresponding to the specified
            # water years, including duplicate indexes for repeating years.
            inds = [
                np.where(self.water_year_array == year)[0] for year in years
            ]
            jx_valid = np.hstack(inds)

        return jx_valid

    def apply(
        self,
        func: Callable[..., ArrayLike],
        reps: int,
        extra_kwargs: dict[str, Any] | None = None,
    ) -> Float64Array:  # type: ignore
        """
        Apply a function to bootstrap replicated data.

        Parameters
        ----------
        func : callable
            Function the computes parameter values.  See Notes for requirements
        reps : int, default 1000
            Number of bootstrap replications
        extra_kwargs : dict, default None
            Extra keyword arguments to use when calling func.  Must not
            conflict with keyword arguments used to initialize bootstrap

        Returns
        -------
        ndarray
            reps by nparam array of computed function values where each row
            corresponds to a bootstrap iteration
        """
        # JAB:
        # if not self.user_defined_boot_years:
        #     self.boot_year_array = np.full(
        #         (self.num_water_years, reps), FILL_VALUE, np.int32
        #     )

        logger.debug(f"Apply the Gumboot bootstrap method over {reps} reps.")

        reps = reps + 1

        kwargs = _add_extra_kwargs(self._kwargs, extra_kwargs)
        base = func(*self._args, **kwargs)
        try:
            num_params = base.shape[0]
        except (IndexError, AttributeError):
            num_params = 1
        results = np.zeros((reps, num_params))
        count = 0
        for pos_data, kw_data in self.bootstrap(reps):
            kwargs = _add_extra_kwargs(kw_data, extra_kwargs)
            results[count] = func(*pos_data, **kwargs)
            count += 1

        # JAB calculations here?

        return results

    def bootstrap(
        self, reps: int
    ) -> PyGenerator[
            tuple[tuple[ArrayLike, ...], dict[str, ArrayLike]],
            None,
            None
    ]:
        """
        Create iterator for use when bootstrapping.

        Parameters
        ----------
        reps : int
            Number of bootstrap replications

        Returns
        -------
        generator
            Generator to iterate over in bootstrap calculations

        Notes
        -----
        The iterator returns a tuple containing the data entered in positional
        arguments as a tuple and the data entered using keywords as a
        dictionary
        """
        for rep in range(reps):
            self._index = self.update_indices(rep)
            yield self._resample()



"""Module contains utility functions."""
from pathlib import Path
import shutil
from typing import Union, Tuple
from teehr.utils.s3path import S3Path
import logging

logger = logging.getLogger(__name__)


def to_path_or_s3path(path: Union[str, Path, S3Path], *args: Tuple[str, ...]) -> Union[Path, S3Path]:
    """Concatenate string(s) to Path or S3Path.

    Check if the path is a string, Path or S3Path.
    If it is a string, check if it is an S3 path (based on beginning of string) and return an S3Path.
    If it is a Path, return a Path.
    If it is an S3Path, return an S3Path.
    Concatenate the strings to the path and return the concatenated path.
    """

    for arg in args:
        if not isinstance(arg, str):
            raise ValueError("Only strings can be concatenated to S3Path or Path")

    if isinstance(path, str):
        # check if s3 path, return an S3Path
        if path.startswith("s3://") or path.startswith("s3a://") or path.startswith("s3n://"):
            path = S3Path(path, *args)
            return path

        # else, try regular path, return a Path
        path = Path(path, *args)
        return path

    if isinstance(path, S3Path):
        path = S3Path(path, *args)
        return path

    # try to treat as path
    path = Path(path, *args)
    return path


def path_to_spark(path: Union[str, Path, S3Path], pattern: str = None) -> str:
    """Convert a Path or S3Path to a string or list of strings."""
    if isinstance(path, S3Path):
        path = str(path)

    if isinstance(path, Path):
        if path.is_dir():
            if pattern is None:
                path = str(path)
            else:
                path = [str(f) for f in path.glob(pattern)]
        else:
            path = str(path)

    return path


def remove_dir_if_exists(path: Union[str, Path]):
    """Remove directory if it exists."""
    if isinstance(path, str):
        path = Path(path)
    if path.is_dir():
        shutil.rmtree(path)
